"""Tests for metabot.util.tickets."""

from metabot.util import tickets


def test_brownpapertickets(requests_mock):
    """Quick tests for _brownpapertickets."""

    assert list(tickets.get_info('')) == []

    requests_mock.get('https://m.bpt.me/event/12345', text='Empty')

    assert list(tickets.get_info('Event http://brownpapertickets.com/event/12345 info')) == [
        (0, 'https://bpt.me/event/12345'),
    ]

    requests_mock.get('https://m.bpt.me/event/12345', text='<TD>789 tickets remaining</TD>')

    assert list(tickets.get_info('Event http://brownpapertickets.com/event/12345 info')) == [
        (789, 'https://bpt.me/event/12345'),
    ]

    assert list(
        tickets.get_info(
            'Event http://brownpapertickets.com/event/12345 info http://m.bpt.me/event/12345 dupe')
    ) == [(789, 'https://bpt.me/event/12345')]

    requests_mock.get('https://m.bpt.me/event/23456', text='<TD>987 tickets remaining</TD>')

    assert list(
        tickets.get_info(
            'Event http://brownpapertickets.com/event/12345 info http://m.bpt.me/event/23456 two')
    ) == [(789, 'https://bpt.me/event/12345'), (987, 'https://bpt.me/event/23456')]


def test_eventbrite(requests_mock):
    """Quick tests for _eventbrite."""

    assert list(tickets.get_info('')) == []

    requests_mock.get('https://www.eventbrite.com/ajax/event/12345/ticket_classes/for_sale/',
                      text='Empty')

    assert list(
        tickets.get_info(
            'Event https://www.eventbrite.com/e/my-cool-event-tickets-12345 info')) == []

    requests_mock.get(
        'https://www.eventbrite.com/ajax/event/12345/ticket_classes/for_sale/',
        json={
            'event_info': {
                'vanity_url': 'https://my-cool_event.eventbrite.com',
                'status_is_sold_out': False,
                'canonical_url': 'https://www.eventbrite.com/e/my-cool-event-tickets-12345',
            },
            'ticket_classes': [{
                'number_of_tickets_remaining': None,
            }],
        })

    assert list(
        tickets.get_info(
            'Event https://www.eventbrite.com/e/my-cool-event-tickets-12345 info')) == [
                (0, 'https://my-cool_event.eventbrite.com'),
            ]

    requests_mock.get(
        'https://www.eventbrite.com/ajax/event/12345/ticket_classes/for_sale/',
        json={
            'event_info': {
                'vanity_url': 'https://my-cool_event.eventbrite.com',
                'status_is_sold_out': False,
                'canonical_url': 'https://www.eventbrite.com/e/my-cool-event-tickets-12345',
            },
            'ticket_classes': [{
                'number_of_tickets_remaining': 789,
            }],
        })

    assert list(
        tickets.get_info(
            'Event https://www.eventbrite.com/e/my-cool-event-tickets-12345 info')) == [
                (789, 'https://my-cool_event.eventbrite.com'),
            ]



"""Tests for metabot.modules.newbot."""

import ntelebot
import pytest

from metabot.modules import newbot


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(newbot)


# pylint: disable=line-too-long


def test_default(conversation):  # pylint: disable=redefined-outer-name
    """Verify the bot's /newbot command."""

    assert conversation.message('/notnewbot') == ''

    assert conversation.message('/newbot') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Add a bot: <b>Paste a bot API Token</b>

To create a new bot, let me know your bot account's API Token. This is a code that looks like:

<pre>123456:ABC-DEF1234ghIkl-zyx57W2v1u123ew11</pre>

If you are creating a bot entirely from scratch, open a private chat with @BotFather and follow the process at https://core.telegram.org/bots#creating-a-new-bot to create a new bot account. At the end of that process, you will receive an API Token that you can paste here.

Otherwise, open a private chat with @BotFather, type <code>/mybots</code>, select the bot account you want to use, select <code>API Token</code>, then copy the code and paste it here:
"""

    assert conversation.message('bogus') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Add a bot: <b>Paste a bot API Token</b>

Oops, <code>bogus</code> doesn't look like an API Token.

To create a new bot, let me know your bot account's API Token. This is a code that looks like:

<pre>123456:ABC-DEF1234ghIkl-zyx57W2v1u123ew11</pre>

If you are creating a bot entirely from scratch, open a private chat with @BotFather and follow the process at https://core.telegram.org/bots#creating-a-new-bot to create a new bot account. At the end of that process, you will receive an API Token that you can paste here.

Otherwise, open a private chat with @BotFather, type <code>/mybots</code>, select the bot account you want to use, select <code>API Token</code>, then copy the code and paste it here:
"""

    ntelebot.bot.Bot('1234:invalid').getme.respond(json={
        'description': 'Unauthorized',
        'error_code': 401,
        'ok': False,
    })

    assert conversation.message('1234:invalid') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Add a bot: <b>Paste a bot API Token</b>

Woops, Telegram told me <code>1234:invalid</code> is unauthorized, meaning the code is either incomplete or out of date. If you generated it yourself, open a private chat with @BotFather, type <code>/mybots</code>, select the bot account you're trying to use, select <code>API Token</code>, then copy the code and paste it here. If you got the code from someone else, send them these instructions (including the token I used). If the code you got from BotFather isn't working, select <code>Revoke current token</code> to generate a new one, then paste that new one here:
"""

    mockbot = ntelebot.bot.Bot('1234:valid')
    mockbot.getme.respond(json={
        'ok': True,
        'result': {
            'first_name': 'Valid Bot',
            'id': 1234,
            'username': 'validbot',
        }
    })
    mockbot.getupdates.respond(json={'description': 'Conflict', 'error_code': 409, 'ok': False})

    assert conversation.message('1234:valid') == """\
[chat_id=1000 parse_mode=HTML]
Cool, that API Token is for <code>validbot</code>. Give me another moment...


[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Add a bot: <b>Paste a bot API Token</b>

Woops, it looks like this bot account is already in use. Make sure no other bot programs are running using this API Token and paste the token again, or use another one:
"""

    mockbot.getme.respond(json={
        'ok': True,
        'result': {
            'first_name': 'Valid Bot',
            'id': 1234,
            'username': 'validbot',
        },
    })
    mockbot.getupdates.respond(json={'description': 'Not Found', 'error_code': 404, 'ok': False})

    assert conversation.message('1234:valid') == """\
[chat_id=1000 parse_mode=HTML]
Cool, that API Token is for <code>validbot</code>. Give me another moment...


[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Add a bot: <b>Paste a bot API Token</b>

Woops, while trying to use <code>1234:valid</code> I got error 404 (<code>Not Found</code>).
"""

    mockbot.getme.respond(json={
        'ok': True,
        'result': {
            'first_name': 'Valid Bot',
            'id': 1234,
            'username': 'validbot',
        },
    })
    mockbot.getupdates.respond(json={'ok': True, 'result': []})

    assert conversation.message('1234:valid') == """\
[chat_id=1000 parse_mode=HTML]
Cool, that API Token is for <code>validbot</code>. Give me another moment...


[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Add a bot: <b>Configure your new bot</b>

Yay, I am now running <code>validbot</code> (<code>Valid Bot</code>). Open a private chat with @validbot and type <code>/admin</code> to continue setup.
"""

    assert conversation.multibot.conf['bots'] == {
        'modulestestbot': {
            'issue37': {
                'admin': {
                    'admins': [1000],
                },
                'telegram': {
                    'running': False,
                    'token': '1234:test',
                },
            },
        },
        'validbot': {
            'issue37': {
                'admin': {
                    'admins': [1000],
                },
                'telegram': {
                    'running': True,
                    'token': '1234:valid',
                },
            },
        },
    }


def test_help(conversation):  # pylint: disable=redefined-outer-name
    """Test /help."""

    assert conversation.message('/help', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Commands</b>

/newbot – Set up a new bot
"""



"""Simple group/supergroup moderator tools."""

import logging
import random

import ntelebot

from metabot.util import adminui
from metabot.util import humanize


def modinit(multibot):  # pylint: disable=missing-docstring

    def _queue():
        multibot.loop.queue.puthourly(30 * 60, _hourly, jitter=random.random() * 5)

    def _hourly():
        try:
            checked = set()
            for botconf in multibot.conf['bots'].values():
                # See https://github.com/nmlorg/metabot/issues/26.
                bot = ntelebot.bot.Bot(botconf['issue37']['telegram']['token'])
                bot.multibot = multibot
                for groupid in botconf['issue37']['moderator']:
                    groupid = int(groupid)
                    if groupid in checked:
                        continue
                    groupdata = multibot.conf['groups'][groupid]
                    try:
                        data = bot.get_chat_administrators(chat_id=groupid)
                    except ntelebot.errors.Error:
                        continue
                    checked.add(groupid)
                    groupdata['admins'] = sorted(member['user']['id'] for member in data)
            log = multibot.conf.finalize()
            if log:
                for path, (value, orig) in sorted(log.items()):
                    pathstr = '.'.join('%s' % part for part in path)
                    logging.info('[moderator] %s: %r -> %r', pathstr, orig, value)
                multibot.conf.save()
        finally:
            _queue()

    _queue()


def modpredispatch(ctx, unused_msg, modconf):  # pylint: disable=missing-docstring
    if ctx.chat and ctx.chat['type'] in ('channel', 'group', 'supergroup'):
        group_id = '%s' % ctx.chat['id']
        groupconf = modconf[group_id]
        groupconf['title'] = ctx.chat.get('title')

        if ctx.type == 'message':
            for target_group_id, target_groupconf in ctx.bot.config['issue37']['moderator'].items():
                if target_groupconf['forward']['from'] == ctx.chat['id']:
                    ctx.forward(int(target_group_id),
                                disable_notification=not target_groupconf['forward']['notify'])


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if ctx.type == 'join':
        join(ctx, msg, modconf)

    if ctx.type in ('message', 'callback_query') and ctx.command == 'mod':
        ctx.private = True
        msg.path('/mod', 'Group Admin')
        ctx.targetbotconf = ctx.bot.config
        return admin(adminui.Frame(ctx, msg, ctx.bot.config, 'moderator', 'Group Admin', ctx.text),
                     botadmin=False)

    return False


def join(ctx, msg, modconf):
    """Respond to new users joining a group chat."""

    greeting = modconf['%s' % ctx.chat['id']].get('greeting')
    if not greeting:
        return
    users = [user for user in ctx.data if not user['is_bot']]
    if not users:
        return

    if 'pinned message' in greeting and ctx.groupinfo.data.pinned_message_id:
        if ctx.groupinfo.data.username:
            url = 'https://t.me/%s/%s' % (ctx.groupinfo.data.username,
                                          ctx.groupinfo.data.pinned_message_id)
        elif -1002147483647 <= ctx.chat['id'] < -1000000000000:
            # See https://github.com/nmlorg/metabot/issues/43.
            url = 'https://t.me/c/%s/%s' % (-1000000000000 - ctx.chat['id'],
                                            ctx.groupinfo.data.pinned_message_id)
        else:
            url = None  # pragma: no cover
        if url:
            greeting = greeting.replace('pinned message', '<a href="%s">pinned message</a>' % url)

    if 'new users' in greeting:
        names = [
            '<a href="tg://user?id=%s">%s</a>' % (user['id'], user['first_name']) for user in users
        ]
        greeting = greeting.replace('new users', humanize.list(names))

    greeting = greeting.replace('chat title', ctx.groupinfo.data.title)

    msg.quiet = True
    msg.add(greeting)


def admin(frame, botadmin=True):  # pylint: disable=too-many-arguments
    """Handle /admin BOTNAME moderator."""

    ctx, msg = frame.ctx, frame.msg
    menu = adminui.Menu()
    for group_id in frame.value:
        groupdata = ctx.bot.multibot.conf['groups'][int(group_id)]
        if botadmin or ctx.user['id'] in groupdata['admins']:
            menu.add(group_id, desc=groupdata['title'])

    if not menu.fields:
        if botadmin:
            msg.action = 'Add me to a group'
            return msg.add(
                "I'm not in any groups! Add me to an existing group from its details screen.")
        msg.action = 'Become a group admin'
        return msg.add(
            "Hi! You aren't an admin in any groups I'm in. If you should be, ask a current admin "
            "to promote you from the group's members list.")

    frame, handler = menu.select(frame)
    if not handler:
        return menu.display(frame, what='group')

    msg.path(frame.field)

    adminui.Menu(
        ('calendars', adminui.calendars, 'Which calendars should be listed in /events?'),
        ('daily', adminui.announcement, 'Should I announce upcoming events once a day?'),
        ('forward', adminui.forward, 'Automatically forward messages from one chat to this one.'),
        ('greeting', adminui.freeform, 'How should I greet people when they join?'),
        ('maxeventscount', adminui.integer, 'How many events should be listed in /events?'),
        ('maxeventsdays', adminui.integer, 'How many days into the future should /events look?'),
        ('timezone', adminui.timezone, 'What time zone should be used in /events?'),
    ).handle(frame)



"""Tests for metabot.util.dicttools."""

from metabot.util import dicttools


def test_audit():
    """Test the audit logging logic."""

    cont = dicttools.ImplicitTrackingDict()
    assert cont.finalize() == {}

    cont['alpha'] = 5
    assert cont.finalize() == {('alpha',): (5, None)}

    cont['alpha'] = 6
    assert cont.finalize() == {('alpha',): (6, 5)}

    cont['alpha'] = 6
    assert cont.log == {}

    cont['alpha'] = 7
    assert cont.log == {('alpha',): (7, 6)}

    cont['alpha'] = 7
    assert cont.log == {('alpha',): (7, 6)}

    cont['alpha'] = 8
    assert cont.log == {('alpha',): (8, 6)}

    cont['alpha'] = 6
    assert cont.log == {}


def test_pop():
    """Test ImplicitTrackingDict.pop."""

    cont = dicttools.ImplicitTrackingDict({'alpha': 'bravo'})
    assert cont.finalize() == {('alpha',): ('bravo', None)}

    assert cont.pop('unset') is None
    assert cont.finalize() == {}

    del cont['unset']
    assert cont.finalize() == {}

    assert cont.pop('alpha') == 'bravo'
    assert cont.finalize() == {('alpha',): (None, 'bravo')}


def test_setitem():
    """Test ImplicitTrackingDict.__setitem__."""

    cont = dicttools.ImplicitTrackingDict()
    assert cont.log == {}

    cont['alpha'] = {}
    assert cont.log == {}
    assert isinstance(cont['alpha'], dicttools.ImplicitTrackingDict)
    assert cont['alpha'].log is cont.log

    cont['alpha']['bravo'] = {}
    assert cont.log == {}
    assert cont['alpha']['bravo'].log is cont.log

    cont['alpha']['bravo']['charlie'] = 5
    cont['alpha']['bravo']['delta'] = 6
    assert cont.finalize() == {
        ('alpha', 'bravo', 'charlie'): (5, None),
        ('alpha', 'bravo', 'delta'): (6, None)
    }

    cont['alpha'] = 7
    assert cont.finalize() == {
        ('alpha', 'bravo', 'charlie'): (None, 5),
        ('alpha', 'bravo', 'delta'): (None, 6),
        ('alpha',): (7, None)
    }

    cont['alpha'] = b'ascii'
    assert cont.finalize() == {('alpha',): ('ascii', 7)}

    cont['alpha'] = None
    assert cont.finalize() == {('alpha',): (None, 'ascii')}

    assert 'alpha' not in cont


def test_missing():
    """Test ImplicitTrackingDict.__missing__."""

    cont = dicttools.ImplicitTrackingDict()

    assert list(cont.items()) == []
    assert 'unset' not in cont
    assert list(iter(cont)) == []

    # __missing__
    unset = cont['unset']
    assert cont['unset'] is unset
    assert unset == {}
    assert isinstance(unset, dicttools.ImplicitTrackingDict)

    assert cont.finalize() == {}

    assert list(cont.items()) == []
    assert 'unset' not in cont
    assert list(iter(cont)) == []


def test_list():
    """Test TrackingList (via ImplicitTrackingDict.__setitem__)."""

    cont = dicttools.ImplicitTrackingDict()

    cont['alpha'] = [2, 4, 6]  # extend
    assert cont.finalize() == {('alpha',): ((2, 4, 6), ())}

    cont['alpha'].append(8)
    assert cont.finalize() == {('alpha',): ((2, 4, 6, 8), (2, 4, 6))}

    cont['alpha'][1] = 10
    assert cont.finalize() == {('alpha',): ((2, 10, 6, 8), (2, 4, 6, 8))}

    cont['alpha'][0] = 2
    assert cont.finalize() == {}

    cont['alpha'][0] = 3
    cont['alpha'][0] = 2
    assert cont.finalize() == {}

    assert cont['alpha'].pop(0) == 2
    assert cont.finalize() == {('alpha',): ((10, 6, 8), (2, 10, 6, 8))}

    cont['alpha'].sort()
    assert cont.finalize() == {('alpha',): ((6, 8, 10), (10, 6, 8))}

    cont['alpha'].remove(8)
    assert cont.finalize() == {('alpha',): ((6, 10), (6, 8, 10))}

    cont['alpha'].reverse()
    assert cont.finalize() == {('alpha',): ((10, 6), (6, 10))}

    cont['alpha'].insert(0, 20)
    assert cont.finalize() == {('alpha',): ((20, 10, 6), (10, 6))}

    cont['alpha'] = [5, 4, 3]
    assert cont.finalize() == {('alpha',): ((5, 4, 3), (20, 10, 6))}

    cont['alpha'].clear()
    assert cont.finalize() == {('alpha',): ((), (5, 4, 3))}



"""Tests for metabot.util.modutil."""

from metabot import conftest
from metabot import multibot
from metabot import test_multibot
from metabot.util import modutil


def test_basic():
    """Quick test for load_modules."""

    modules = modutil.load_modules('metabot')
    assert conftest not in modules
    assert multibot in modules
    assert test_multibot not in modules



"""Tests for metabot.util.html."""

from metabot.util import html


def test_sanitize():
    """Test sanitization logic."""

    assert html.sanitize('') == ''
    assert html.sanitize('empty') == 'empty'
    assert html.sanitize('&#8592; &#x2190; &larr; \u2190') == '← ← ← ←'
    assert html.sanitize('&#38; &#x26; &amp;') == '&amp; &amp; &amp;'

    # From https://core.telegram.org/bots/api#formatting-options.
    text = """
<b>bold</b>, <strong>bold</strong>
<i>italic</i>, <em>italic</em>
<a href="http://www.example.com/">inline URL</a>
<a href="tg://user?id=123456789">inline mention of a user</a>
<code>inline fixed-width code</code>
<pre>pre-formatted fixed-width code block</pre>
"""
    assert html.sanitize(text) == text
    assert html.sanitize('a<br>b<div>c</div><p>d</p>') == 'a\nb\nc\nd'
    assert html.sanitize('<a weird="true">text</a>') == '<a>text</a>'

    assert html.sanitize('<b>good</b> <span>bad</span>') == '<b>good</b> bad'
    assert html.sanitize('<b><span>nested</span></b>') == '<b>nested</b>'
    assert html.sanitize('<span><b>nested</b></span>') == '<b>nested</b>'
    assert html.sanitize('<b>broken <i>good</i>') == '<b>broken </b><i>good</i>'

    assert html.sanitize('&bogus;') == '&amp;bogus;'
    assert html.sanitize('&#bogus;') == '&amp;#bogus;'
    assert html.sanitize('&#xxbogus;') == '&amp;#xxbogus;'


def test_strip():
    """Test stripping logic."""

    assert html.sanitize('', True) == ''
    assert html.sanitize('empty', True) == 'empty'
    assert html.sanitize('&#8592; &#x2190; &larr; \u2190', True) == '← ← ← ←'
    assert html.sanitize('&#38; &#x26; &amp;', True) == '&amp; &amp; &amp;'

    # From https://core.telegram.org/bots/api#formatting-options.
    text = """
<b>bold</b>, <strong>bold</strong>
<i>italic</i>, <em>italic</em>
<a href="http://www.example.com/">inline URL</a>
<a href="tg://user?id=123456789">inline mention of a user</a>
<code>inline fixed-width code</code>
<pre>pre-formatted fixed-width code block</pre>
"""
    assert html.sanitize(text, True) == """
bold, bold
italic, italic
inline URL
inline mention of a user
inline fixed-width code
pre-formatted fixed-width code block
"""
    assert html.sanitize('first<br>second', True) == 'first\nsecond'
    assert html.sanitize('<a weird="true">text</a>', True) == 'text'

    assert html.sanitize('<b>good</b> <span>bad</span>', True) == 'good bad'
    assert html.sanitize('<b><span>nested</span></b>', True) == 'nested'
    assert html.sanitize('<span><b>nested</b></span>', True) == 'nested'
    assert html.sanitize('<b>broken <i>good</i>', True) == 'broken good'

    assert html.sanitize('&bogus;', True) == '&amp;bogus;'
    assert html.sanitize('&#bogus;', True) == '&amp;#bogus;'
    assert html.sanitize('&#xxbogus;', True) == '&amp;#xxbogus;'



"""Tests for metabot.util.jsonutil."""

from metabot.util import jsonutil


def test_load_nonexistent(tmpdir):
    """Verify the loader silently ignores nonexistent files."""

    tmpfile = tmpdir.join('test.json')
    assert jsonutil.load(tmpfile.strpath) is None


def test_load_empty(tmpdir):
    """Verify the loader silently ignores empty files."""

    tmpfile = tmpdir.join('test.json')
    tmpfile.write('')
    assert jsonutil.load(tmpfile.strpath) is None


def test_load_malformed(tmpdir):
    """Verify the loader silently ignores malformed files."""

    tmpfile = tmpdir.join('test.json')
    tmpfile.write('bogus data')
    assert jsonutil.load(tmpfile.strpath) is None


def test_dump_simple(tmpdir):
    """Verify the dumper preserves simple objects."""

    tmpfile = tmpdir.join('test.json')
    obj = 123
    assert jsonutil.dump(tmpfile.strpath, obj) == obj
    assert tmpfile.read() == '123'


def test_dump_load(tmpdir):
    """Verify the dumper and loader work as expected."""

    tmpfile = tmpdir.join('test.json')
    obj = {'key': ['value', 'value']}
    assert jsonutil.dump(tmpfile.strpath, obj) == obj
    assert jsonutil.load(tmpfile.strpath) == obj



"""Quick template for metabot's custom message style."""


def cgi_escape(text):  # pylint: disable=missing-docstring
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>',
                                                                   '&gt;').replace('"', '&quot;')


class MessageBuilder:
    """Quick template for metabot's custom message style."""

    action = None
    quiet = False

    def __init__(self):
        self._path = []
        self._title = []
        self._lines = []
        self._keyboard = []

    def __bool__(self):
        return bool(self._title or self._lines)

    __nonzero__ = __bool__

    def path(self, path, title=None):
        """Add path to the visible title and to the path used to calculate button callbacks."""

        self._path.append(path)
        self._title.append(title or path)

    def pathpop(self):
        """Remove the most-recently added path/title component."""

        self._path.pop()
        self._title.pop()

    def add(self, line, *args):
        """Add line to the message body, interpolating HTML-escaped args if provided."""

        if args:
            line %= tuple(cgi_escape(str(arg)) for arg in args)
        self._lines.append(line)

    def _make_button(self, data):
        if data:
            text, callback_data = data
        else:
            text = '\xa0'
            callback_data = '/stop'
        path = self._path[:]
        while path and callback_data.startswith('..'):
            path.pop()
            callback_data = callback_data[len('..'):].lstrip()
        if not callback_data.startswith('/'):
            if callback_data:
                path.append(callback_data)
            callback_data = ' '.join(path)
        return {'text': cgi_escape(text), 'callback_data': callback_data}

    def button(self, text, callback_data):
        """Add a row to the keyboard with single callback button."""

        self.buttons([(text, callback_data)])

    def buttons(self, buttons):
        """Add a row to the keyboard with one or more callback buttons."""

        self._keyboard.append(buttons)

    def reply(self, ctx):
        """Build the message and pipe it through ctx.reply_html."""

        first_line = cgi_escape(' \u203a '.join(self._title))
        if first_line and self.action:
            first_line += ': '
        if self.action:
            first_line = '%s<b>%s</b>' % (first_line, cgi_escape(self.action))
        lines = first_line and [first_line] + self._lines or self._lines
        if len(self._path) > 1:
            self.button('Back', '..')
        kwargs = {}
        if self._keyboard:
            keyboard = [[self._make_button(data) for data in row] for row in self._keyboard]
            kwargs['reply_markup'] = {'inline_keyboard': keyboard}
        if self.quiet:
            kwargs['disable_notification'] = True
        ctx.set_conversation(' '.join(self._path))
        return ctx.reply_html('\n\n'.join(lines), disable_web_page_preview=True, **kwargs)



"""Simple static calendar for use in tests."""

from metabot.calendars import base


class Calendar(base.Calendar):
    """Simple static calendar for use in tests."""

    caltype = 'static'
    __last_id = 0
    __last_update = 0
    poll_result = False

    def __init__(self, calid):
        super().__init__(calid)
        self.__live_events = {}

    def poll(self):  # pylint: disable=arguments-differ
        return self.poll_result

    @staticmethod
    def event_proto_to_local(proto):
        return {key.lower(): value for key, value in proto.items()}

    @staticmethod
    def event_local_to_proto(local):
        return {key.upper(): value for key, value in local.items()}

    def proto_add(self, proto):  # pylint: disable=arguments-differ
        newproto = dict(proto)
        self.__last_id += 1
        newproto['ID'] = '%s' % self.__last_id
        self.__last_update += 1
        newproto['UPDATED'] = 1000 + self.__last_update
        self.__live_events[newproto['ID']] = newproto
        return newproto

    def proto_update(self, proto_id, proto):  # pylint: disable=arguments-differ
        current = self.__live_events[proto_id]
        current.update(proto)
        self.__last_update += 1
        current['UPDATED'] = 1000 + self.__last_update
        return current

    def proto_remove(self, proto_id):  # pylint: disable=arguments-differ
        self.__live_events.pop(proto_id)



"""Reusable UI elements for admin functions (like /admin BOTNAME moderator and /events set)."""

import math

import pytz

from metabot.util import humanize
from metabot.util import tzutil


class Frame:
    """The current position (and related state) within a hierarchical message."""

    def __init__(self, ctx, msg, parent, field, desc, text):  # pylint: disable=too-many-arguments
        self.ctx = ctx
        self.msg = msg
        self.parent = parent.get('issue37') or parent
        self.field = field
        self.desc = desc
        self.text = text

    def get(self, default=None):
        """Return the current value iff it already exists."""

        return self.parent.get(self.field, default)

    @property
    def value(self):
        """Return the current value if it already exists, otherwise create it as a new container."""

        return self.parent[self.field]

    @value.setter
    def value(self, value):
        self.parent[self.field] = value


class Menu:
    """Manage the UX of navigating into each Frame."""

    def __init__(self, *fields):
        self.fields = []
        for field in fields:
            self.add(*field)

    def add(self, field, handler=True, desc=None):
        """Add a possible subframe."""

        self.fields.append((field, handler, desc))
        self.fields.sort(key=lambda field: (field[0] is None, field[0]))

    def select(self, frame):
        """Identify the selected subframe."""

        field, _, text = frame.text.lstrip().partition(' ')
        if field:
            field = field.lower()
            for fieldname, handler, desc in self.fields:
                if fieldname is None:
                    fieldname = field
                if fieldname.lower() == field:
                    return Frame(frame.ctx, frame.msg, frame.value, fieldname, desc, text), handler
            frame.msg.add("I can't set <code>%s</code>.", field)
        return frame, None

    def dispatch(self, frame):
        """Enter the selected subframe if possible."""

        frame, handler = self.select(frame)
        if handler:
            frame.msg.path(frame.field)
            handler(frame)
            if frame.msg.action:
                return True
            frame.msg.pathpop()

    def display(self, frame, what='field'):
        """Display the current available subframes as a menu."""

        msg = frame.msg
        msg.action = 'Choose a ' + what
        for field, handler, desc in self.fields:
            if field is None:
                continue
            value = frame.value.get(field)
            if handler is bool:
                value = value and 'yes' or 'no'
            elif isinstance(value, dict):
                value = None
            elif value is not None:
                value = '%s' % value
                if len(value) > 10:
                    value = value[:9] + '\u2026'
            label = field
            if value:
                label = '%s (%s)' % (label, value)
            if desc:
                label = '%s \u2022 %s' % (label, desc)
            msg.button(label, field)

    def handle(self, frame, **kwargs):
        """Enter the selected subframe if possible, or display the menu."""

        if not self.dispatch(frame):
            self.display(frame, **kwargs)


def announcement(frame):
    """Configure a daily announcement."""

    menu = Menu(
        ('hour', integer, 'At what hour?'),
        ('dow', daysofweek, 'Which days of the week should I announce upcoming events on?'),
        ('text', freeform,
         'One or more messages (one per line) to use/cycle through for the daily announcement.'),
    )
    if not menu.dispatch(frame):
        frame.msg.add(frame.desc)
        menu.display(frame)


def bool(frame):  # pylint: disable=redefined-builtin
    """Configure a toggle-able setting."""

    if frame.get():
        frame.value = None
        frame.msg.add('Disabled <code>%s</code>.', frame.field)
    else:
        frame.value = True
        frame.msg.add('Enabled <code>%s</code>.', frame.field)


def calendars(frame):
    """Configure a selection of calendars."""

    ctx, msg = frame.ctx, frame.msg
    action, _, target = frame.text.partition(' ')

    calcodes = set(frame.get('').split())

    if target and target not in ctx.bot.multibot.calendars:
        msg.add('<code>%s</code> is not a calendar!', target)
    elif action == 'add' and target:
        if target in calcodes:
            msg.add('<code>%s</code> is already in your calendar view!', target)
        else:
            msg.add('Added <code>%s</code> to your calendar view.', target)
            calcodes.add(target)
    elif action == 'remove' and target:
        if target not in calcodes:
            msg.add('<code>%s</code> is not in your calendar view!', target)
        else:
            msg.add('Removed <code>%s</code> from your calendar view.', target)
            calcodes.remove(target)

    frame.value = ' '.join(sorted(calcodes)) or None

    msg.action = 'Select a calendar'
    msg.add(frame.desc)
    msg.add('Select a calendar to add or remove from the list below:')
    for calcode, calendar_info in sorted(ctx.bot.multibot.calendars.items(),
                                         key=lambda pair: pair[1]['name']):
        if calcode not in calcodes:
            msg.button('Add %s' % calendar_info['name'], 'add %s' % calcode)
        else:
            msg.button('Remove %s' % calendar_info['name'], 'remove %s' % calcode)


def daysofweek(frame):  # pylint: disable=too-many-branches
    """Select days of the week to enable/disable."""

    msg = frame.msg
    value = frame.get(0)
    if frame.text == 'all':
        value = 0
    elif frame.text == 'none':
        value = 127
    elif frame.text.isdigit():
        value ^= 1 << int(frame.text)
    frame.value = value or None

    msg.action = 'Select a day of the week to toggle'
    msg.add(frame.desc)
    if value == 127:
        msg.add('All days are currently <b>disabled</b>.')
    elif not value:
        msg.add('All days are currently <b>enabled</b>.')
    else:
        days = [
            name for name, code in (('Sunday', 6), ('Monday', 0), ('Tuesday', 1), ('Wednesday', 2),
                                    ('Thursday', 3), ('Friday', 4), ('Saturday', 5))
            if not value & 1 << code
        ]
        msg.add('Enabled for %s.', humanize.list(days))

    msg.add('Select a day of the week to toggle:')
    layout = (
        (('Sunday', 6), ('Monday', 0)),
        (('Tuesday', 1), ('Wednesday', 2)),
        (('Thursday', 3), ('Friday', 4)),
        (('Saturday', 5), ('every day', -1)),
    )
    for row in layout:
        buttons = []
        for title, code in row:
            if code == -1:
                if value:
                    buttons.append(('enable ' + title, 'all'))
                else:
                    buttons.append(('disable ' + title, 'none'))
            elif value & 1 << code:
                buttons.append(('enable ' + title, '%s' % code))
            else:
                buttons.append(('disable ' + title, '%s' % code))
        msg.buttons(buttons)


def forward(frame):
    """Configure the bot to forward messages from one chat to another."""

    menu = Menu(
        ('from', groupid, 'What group should messages be forwarded from?'),
        ('notify', bool, 'Should forwarded messages trigger a notification?'),
    )
    if not menu.dispatch(frame):
        frame.msg.add(frame.desc)
        menu.display(frame)


def freeform(frame):  # pylint: disable=too-many-branches
    """Configure a free-form text field."""

    ctx, msg = frame.ctx, frame.msg
    if ctx.document:  # pragma: no cover
        text = 'document:%s' % ctx.document
    elif ctx.photo:  # pragma: no cover
        text = 'photo:%s' % ctx.photo
    elif ctx.sticker:  # pragma: no cover
        text = 'sticker:%s' % ctx.sticker
    else:
        text = frame.text

    if text:
        if text.lower() in ('-', 'none', 'off'):
            text = None
        set_log(frame, text)
    else:
        msg.action = 'Type a new value for ' + frame.field
        msg.add(frame.desc)
        if frame.get():
            msg.add('<code>%s</code> is currently <code>%s</code>.', frame.field, frame.value)
        msg.add('Type your new value, or type "off" to disable/reset to default.')


def groupid(frame):
    """Select a group."""

    ctx, msg = frame.ctx, frame.msg
    if frame.text in ctx.targetbotconf['issue37']['moderator']:
        return set_log(frame, frame.text)

    msg.action = 'Select a group'
    msg.add(frame.desc)
    msg.add('Select a group:')
    for group_id, groupconf in sorted(ctx.targetbotconf['issue37']['moderator'].items()):
        msg.button('%s (%s)' % (group_id, groupconf['title']), group_id)


def image(frame):
    """Configure an image field."""

    ctx, msg = frame.ctx, frame.msg
    if ctx.photo:
        set_log(frame, ctx.photo)
    elif frame.text.lower() in ('-', 'none', 'off'):
        set_log(frame, None)
    elif frame.text.startswith('https://'):
        set_log(frame, frame.text)
    else:
        msg.action = 'Send a new image for ' + frame.field
        if frame.get():
            msg.add('<code>%s</code> is currently <code>%s</code>.', frame.field, frame.value)
        msg.add('Send a new image, or type "off" to disable/reset to default.')


def integer(frame):
    """Configure an integer field."""

    msg = frame.msg
    if frame.text:
        try:
            value = int(frame.text)
        except ValueError:
            value = None
        set_log(frame, value)
    else:
        msg.action = 'Type a new value for ' + frame.field
        msg.add(frame.desc)
        if frame.get():
            msg.add('<code>%s</code> is currently <code>%s</code>.', frame.field, frame.value)
        msg.add('Type your new value, or type "off" to disable/reset to default.')


def set_log(frame, new):
    """Set frame.value = new, and report the change to the user."""

    msg = frame.msg
    current = frame.get()
    frame.value = new
    if current is None and new is None:
        msg.add('<code>%s</code> is already unset.', frame.field)
    elif current is None:
        msg.add('Set <code>%s</code> to <code>%s</code>.', frame.field, new)
    elif new is None:
        msg.add('Unset <code>%s</code> (was <code>%s</code>).', frame.field, current)
    else:
        msg.add('Changed <code>%s</code> from <code>%s</code> to <code>%s</code>.', frame.field,
                current, new)


def timezone(frame):
    """Configure a time zone."""

    msg = frame.msg
    if frame.text in pytz.all_timezones_set:
        return set_log(frame, frame.text)

    country, _, page = frame.text.partition(' ')
    country = country.upper()
    page = page.isdigit() and int(page) or 0

    timezones = [(ent.zone, ent.comment) for ent in tzutil.ZONE_TAB if ent.code == country]

    if not timezones:
        msg.action = 'Type your 2-letter country code'
        if country:
            msg.add('Unknown country code <code>%s</code>.', country)
        return msg.add('Type your 2-letter country code (like US, CA, GB, etc.).')

    pages = math.ceil(len(timezones) / 7)
    timezones = timezones[page * 7:(page + 1) * 7]

    msg.action = 'Choose a primary city'
    msg.add(frame.desc)
    msg.add('Choose a primary city:')
    for tzname, comment in timezones:
        title = tzname.rsplit('/', 1)[-1].replace('_', ' ')
        if comment:
            title = '%s (%s)' % (title, comment)
        msg.button(title, tzname)
    buttons = [None, None]
    if page:
        buttons[0] = ('Prev', '%s %i' % (country, page - 1))
    if page < pages - 1:
        buttons[1] = ('Next', '%s %i' % (country, page + 1))
    if buttons[0] or buttons[1]:
        msg.buttons(buttons)



"""Tests for metabot.modules.echo."""

import pytest

from metabot.modules import echo


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(echo)


# pylint: disable=line-too-long


def test_echo(conversation):  # pylint: disable=redefined-outer-name
    """Verify the echo module (which uses dynamic commands)."""

    assert conversation.message('/myecho') == ''

    conversation.bot.config['issue37']['echo']['myecho'] = {
        'text': 'These are the rules: Have fun!',
    }

    assert conversation.message('/myecho') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
These are the rules: Have fun!
"""

    conversation.bot.config['issue37']['echo']['about'] = {
        'text': ('First line.\n'
                 'Second line.\n'
                 ' \n'
                 'Last line.'),
    }

    assert conversation.message('/about') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
First line.
Second line.
 
Last line.
"""

    conversation.bot.config['issue37']['echo']['about'] = {
        'text': ('First line.\n'
                 'Second line.\n'
                 ' \n'
                 'Last line.'),
        'paginate': True,
    }

    assert conversation.message('/about') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
First line.
[More (1/3) | /about 2]
"""

    assert conversation.message('/about 2') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
First line.

Second line.
[More (2/3) | /about 3]
"""

    assert conversation.message('/about 3') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
First line.

Second line.

Last line.
"""

    assert conversation.message('/about 1000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
First line.

Second line.

Last line.
"""

    assert conversation.message('/about bogus') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
First line.
[More (1/3) | /about 2]
"""


def test_help(conversation):  # pylint: disable=redefined-outer-name
    """Test /help."""

    conversation.bot.config['issue37']['echo']['rules1'] = {
        'text': 'These are the rules: Have fun!',
    }
    conversation.bot.config['issue37']['echo']['rules2'] = {
        'text': 'These are the rules: Have fun!!',
    }

    assert conversation.message('/help', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Commands</b>

/rules1 – &quot;These are the rules: Have fun!&quot;

/rules2 – &quot;These are the rules: Have fun…&quot;
"""


def test_admin(conversation):  # pylint: disable=redefined-outer-name
    """Test /admin BOTNAME echo."""

    assert conversation.message('/admin modulestestbot echo') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo: <b>Choose a command</b>

Type the name of a command to add (like <code>rules</code>—don't include a slash at the beginning!), or select an existing echo.
[Back | /admin modulestestbot]
"""

    assert conversation.message('EchoTest') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › echotest: <b>Choose a field</b>
[paginate (no) • For multiline messages, display just one line at a time? | /admin modulestestbot echo echotest paginate]
[private (no) • Send the message in group chats, or just in private? | /admin modulestestbot echo echotest private]
[text • The message, sticker, or image to send in response to /echotest. | /admin modulestestbot echo echotest text]
[Back | /admin modulestestbot echo]
"""

    assert conversation.message('text') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › echotest › text: <b>Type a new value for text</b>

The message, sticker, or image to send in response to /echotest.

Type your new value, or type "off" to disable/reset to default.
[Back | /admin modulestestbot echo echotest]
"""

    assert conversation.message('my message') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › echotest: <b>Choose a field</b>

Set <code>text</code> to <code>my message</code>.
[paginate (no) • For multiline messages, display just one line at a time? | /admin modulestestbot echo echotest paginate]
[private (no) • Send the message in group chats, or just in private? | /admin modulestestbot echo echotest private]
[text (my message) • The message, sticker, or image to send in response to /echotest. | /admin modulestestbot echo echotest text]
[Back | /admin modulestestbot echo]
"""

    assert conversation.message('/admin modulestestbot echo echotest text new message') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › echotest: <b>Choose a field</b>

Changed <code>text</code> from <code>my message</code> to <code>new message</code>.
[paginate (no) • For multiline messages, display just one line at a time? | /admin modulestestbot echo echotest paginate]
[private (no) • Send the message in group chats, or just in private? | /admin modulestestbot echo echotest private]
[text (new messa…) • The message, sticker, or image to send in response to /echotest. | /admin modulestestbot echo echotest text]
[Back | /admin modulestestbot echo]
"""

    assert conversation.message('/admin modulestestbot echo') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo: <b>Choose a command</b>

Type the name of a command to add (like <code>rules</code>—don't include a slash at the beginning!), or select an existing echo.
[echotest • new message | /admin modulestestbot echo echotest]
[Back | /admin modulestestbot]
"""

    assert conversation.message('textless paginate') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo › textless: <b>Choose a field</b>

Enabled <code>paginate</code>.
[paginate (yes) • For multiline messages, display just one line at a time? | /admin modulestestbot echo textless paginate]
[private (no) • Send the message in group chats, or just in private? | /admin modulestestbot echo textless private]
[text • The message, sticker, or image to send in response to /textless. | /admin modulestestbot echo textless text]
[Back | /admin modulestestbot echo]
"""

    assert conversation.message('/admin modulestestbot echo') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › echo: <b>Choose a command</b>

Type the name of a command to add (like <code>rules</code>—don't include a slash at the beginning!), or select an existing echo.
[echotest • new message | /admin modulestestbot echo echotest]
[textless | /admin modulestestbot echo textless]
[Back | /admin modulestestbot]
"""



"""Manage the bot's Telegram state."""


def admin(frame):
    """Handle /admin BOTNAME telegram (manage the bot's Telegram state)."""

    ctx, msg, modconf = frame.ctx, frame.msg, frame.value
    username = ctx.targetbotuser
    msg.action = 'Choose an action'

    if frame.text == 'stop':
        if not modconf['running']:
            msg.add('@%s is not currently running.', username)
        else:
            ctx.bot.multibot.stop_bot(username)
            msg.add('@%s is now offline.', username)
    elif frame.text == 'start':
        if modconf['running']:
            msg.add('@%s is already running.', username)
        else:
            ctx.bot.multibot.run_bot(username)
            msg.add('@%s is now running.', username)

    if modconf['running']:
        msg.button('Stop bot', 'stop')
    else:
        msg.button('Start bot', 'start')



"""Simplified interface to https://pyyaml.org/wiki/PyYAMLDocumentation."""

import yaml


def load(fname):
    """Load fname as a YAML file, silently returning None on any error."""

    try:
        data = open(fname, 'rb').read()
    except IOError:
        return

    try:
        return yaml.safe_load(data)
    except yaml.error.YAMLError:
        pass


def dump(fname, obj):
    """Save obj as a YAML file to fname."""

    data = yaml.dump_all([obj], indent=4, width=200, Dumper=_SimplifyingDumper).encode('ascii')
    with open(fname, 'wb') as fobj:
        fobj.write(data)
    return obj


class _SimplifyingRepresenter(yaml.representer.SafeRepresenter):
    yaml_multi_representers = {
        dict: yaml.representer.SafeRepresenter.represent_dict,
        list: yaml.representer.SafeRepresenter.represent_list,
    }


# pylint: disable=too-many-ancestors
class _SimplifyingDumper(yaml.emitter.Emitter, yaml.serializer.Serializer, _SimplifyingRepresenter,
                         yaml.resolver.Resolver):

    # pylint: disable=too-many-arguments,too-many-locals
    def __init__(self,
                 stream,
                 default_style=None,
                 default_flow_style=False,
                 canonical=None,
                 indent=None,
                 width=None,
                 allow_unicode=None,
                 line_break=None,
                 encoding=None,
                 explicit_start=None,
                 explicit_end=None,
                 version=None,
                 tags=None,
                 sort_keys=True):
        yaml.emitter.Emitter.__init__(self,
                                      stream,
                                      canonical=canonical,
                                      indent=indent,
                                      width=width,
                                      allow_unicode=allow_unicode,
                                      line_break=line_break)
        yaml.serializer.Serializer.__init__(self,
                                            encoding=encoding,
                                            explicit_start=explicit_start,
                                            explicit_end=explicit_end,
                                            version=version,
                                            tags=tags)
        _SimplifyingRepresenter.__init__(self,
                                         default_style=default_style,
                                         default_flow_style=default_flow_style,
                                         sort_keys=sort_keys)
        yaml.resolver.Resolver.__init__(self)



"""Tests for metabot.util.protobuf."""

from metabot.util import protobuf


def test_get_varint():
    """Test basic behavior of get_varint."""

    # https://developers.google.com/protocol-buffers/docs/encoding#varints
    data = [0b00000001]
    assert (protobuf.get_varint(data), data) == (1, [])

    data = [0b10101100, 0b00000010]
    assert (protobuf.get_varint(data), data) == (300, [])


def test_get_field():
    """Test basic behavior of get_field."""

    # https://developers.google.com/protocol-buffers/docs/encoding#simple
    data = [0x08, 0x96, 0x01]
    assert (protobuf.get_field(data), data) == ((1, 150), [])

    # https://developers.google.com/protocol-buffers/docs/encoding#strings
    data = [0x12, 0x07, 0x74, 0x65, 0x73, 0x74, 0x69, 0x6e, 0x67]
    assert (protobuf.get_field(data), data) == ((2, list(b'testing')), [])


class PBTest1(protobuf.ProtoBuf):  # pylint: disable=missing-docstring,too-few-public-methods
    # https://developers.google.com/protocol-buffers/docs/encoding#simple
    _fields = ((1, 'a', int),)


class PBTest2(protobuf.ProtoBuf):  # pylint: disable=missing-docstring,too-few-public-methods
    # https://developers.google.com/protocol-buffers/docs/encoding#strings
    _fields = ((2, 'b', str),)


class PBTest3(protobuf.ProtoBuf):  # pylint: disable=missing-docstring,too-few-public-methods
    # https://developers.google.com/protocol-buffers/docs/encoding#embedded
    _fields = ((3, 'c', PBTest1),)


def test_protobuf():
    """Test basic behavior of ProtoBuf."""

    pbuf = PBTest3([0x1a, 0x03, 0x08, 0x96, 0x01])
    assert isinstance(pbuf.c, PBTest1)  # pylint: disable=no-member
    assert pbuf.c.a == 150  # pylint: disable=no-member

    pbuf = PBTest2([0x12, 0x07, 0x74, 0x65, 0x73, 0x74, 0x69, 0x6e, 0x67])
    assert pbuf.b == 'testing'  # pylint: disable=no-member



"""Announce upcoming events once a day."""

import datetime
import logging
import operator
import random
import re
import time

import ntelebot

from metabot.util import eventutil
from metabot.util import html
from metabot.util import humanize
from metabot.util import pickleutil


def modinit(multibot):  # pylint: disable=missing-docstring

    def _queue():
        multibot.loop.queue.puthourly(0, _hourly, jitter=random.random() * 5)

    if multibot.conf.confdir:
        recordsfname = multibot.conf.confdir + '/daily.pickle'
        records = pickleutil.load(recordsfname) or {}
    else:
        recordsfname = None
        records = {}

    def _hourly():
        try:
            multibot.multical.poll()
            _daily_messages(multibot, records)
            if recordsfname:
                pickleutil.dump(recordsfname, records)
        finally:
            _queue()

    _queue()


def _daily_messages(multibot, records):  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    now = time.time()
    for botuser, botconf in multibot.conf['bots'].items():  # pylint: disable=too-many-nested-blocks
        for groupid, groupconf in botconf['issue37']['moderator'].items():
            calcodes, tzinfo, count, days, hour, dow = eventutil.get_group_conf(groupconf)
            if not tzinfo or not isinstance(hour, int):
                continue
            # See https://github.com/nmlorg/metabot/issues/26.
            bot = ntelebot.bot.Bot(botconf['issue37']['telegram']['token'])
            bot.multibot = multibot
            form = lambda event: eventutil.format_event(bot, event, tzinfo, full=False)  # pylint: disable=cell-var-from-loop
            title = lambda event: '  \u2022 ' + event['summary']
            nowdt = datetime.datetime.fromtimestamp(now, tzinfo)
            key = (botuser, groupid)
            if nowdt.hour == hour and not dow & 1 << nowdt.weekday():
                events, alerts = eventutil.get_group_events(bot, calcodes, tzinfo, count, days, now)
                _handle_alerts(bot, records, groupid, alerts)
                if events:
                    preambles = groupconf['daily'].get('text', '').splitlines()
                    preamble = (preambles and preambles[nowdt.toordinal() % len(preambles)] or '')
                    text = _format_daily_message(preamble, list(map(form, events)))
                    url = eventutil.get_image(events[0], botconf)
                    message = None
                    if url:
                        try:
                            message = bot.send_photo(chat_id=groupid,
                                                     photo=url,
                                                     caption=text,
                                                     parse_mode='HTML',
                                                     disable_notification=True)
                        except ntelebot.errors.Error:
                            logging.exception('Downgrading to plain text:')
                    if not message:
                        try:
                            message = bot.send_message(chat_id=groupid,
                                                       text=text,
                                                       parse_mode='HTML',
                                                       disable_web_page_preview=True,
                                                       disable_notification=True)
                        except ntelebot.errors.Error:
                            logging.exception('While sending to %s:\n%s', groupid, text)
                    if message:
                        records[key] = (now, [event.copy() for event in events], message)
            elif key in records:
                lastnow, lastevents, lastmessage = records[key]
                lastmap = {event['local_id']: event for event in lastevents}
                events, alerts = eventutil.get_group_events(bot, calcodes, tzinfo, count, days,
                                                            lastnow)
                _handle_alerts(bot, records, groupid, alerts)
                curmap = {event['local_id']: event for event in events}
                bothevents = events.copy()
                bothevents.extend(event for event in lastevents if event['local_id'] not in curmap)
                bothevents.sort(key=operator.itemgetter('start', 'end', 'summary', 'local_id'))
                edits = []
                for event in bothevents:
                    lastevent = lastmap.get(event['local_id'])
                    if not lastevent:
                        edits.append(title(event))
                        edits.append('    \u25e6 New event!')
                        continue
                    event = multibot.multical.get_event(event['local_id'])[1]
                    if not event:
                        edits.append(title(lastevent))
                        edits.append('    \u25e6 Removed.')
                        continue
                    pairs = (
                        (lastevent['summary'], event['summary']),
                        (eventutil.humanize_range(lastevent['start'], lastevent['end'], tzinfo),
                         eventutil.humanize_range(event['start'], event['end'], tzinfo)),
                        (lastevent['location'], event['location']),
                        (html.sanitize(lastevent['description'], strip=True),
                         html.sanitize(event['description'], strip=True)),
                    )  # yapf: disable
                    pieces = []
                    for left, right in pairs:
                        diff = _quick_diff(left, right)
                        if diff:
                            pieces.append(diff)
                    if pieces:
                        edits.append(title(event))
                        for left, right in pieces:
                            edits.append(f'      <s>{left}</s>')
                            edits.append(f'      {right}')

                if not edits:
                    continue
                text = 'Updated:\n' + '\n'.join(edits)
                try:
                    message = bot.send_message(chat_id=groupid,
                                               reply_to_message_id=lastmessage['message_id'],
                                               text=text,
                                               parse_mode='HTML',
                                               disable_web_page_preview=True,
                                               disable_notification=True)
                except ntelebot.errors.Error:
                    logging.exception('While sending to %s:\n%s', groupid, text)
                    continue

                lastnowdt = datetime.datetime.fromtimestamp(lastnow, tzinfo)
                preambles = groupconf['daily'].get('text', '').splitlines()
                preamble = preambles and preambles[lastnowdt.toordinal() % len(preambles)] or ''
                updated = 'Updated ' + humanize.time(nowdt)
                groupidnum = int(groupid)
                if -1002147483647 <= groupidnum < -1000000000000:
                    updated = '<a href="https://t.me/c/%s/%s">%s</a>' % (
                        -1000000000000 - groupidnum, message['message_id'], updated)
                else:
                    updated = '%s (%s)' % (updated, message['message_id'])
                text = '%s\n\n[%s]' % (_format_daily_message(preamble, list(map(form,
                                                                                events))), updated)
                try:
                    if lastmessage.get('caption'):
                        message = bot.edit_message_caption(chat_id=groupid,
                                                           message_id=lastmessage['message_id'],
                                                           caption=text,
                                                           parse_mode='HTML')
                    else:
                        message = bot.edit_message_text(chat_id=groupid,
                                                        message_id=lastmessage['message_id'],
                                                        text=text,
                                                        parse_mode='HTML',
                                                        disable_web_page_preview=True)
                except ntelebot.errors.Error:
                    logging.exception('While sending to %s:\n%s', groupid, text)
                else:
                    records[key] = (lastnow, [event.copy() for event in events], message)


def _quick_diff(left, right):
    left = re.sub(r'https?://\S+', '', left)
    left = re.sub(r'\s+', ' ', left).strip()
    right = re.sub(r'https?://\S+', '', right)
    right = re.sub(r'\s+', ' ', right).strip()
    if left == right:
        return
    i = 0
    while i < len(left) and i < len(right) and right[i] == left[i]:
        i += 1
    if i > 10:
        left = '\u2026' + left[i - 9:]
        right = '\u2026' + right[i - 9:]
    if len(left) > 40:
        left = left[:39] + '\u2026'
    if len(right) > 40:
        right = right[:39] + '\u2026'
    return left, right


def _format_daily_message(preamble, events):
    if len(events) == 1:
        text = "There's an event coming up:"
    elif len(events) == 2:
        text = 'There are a couple events coming up:'
    elif len(events) == 3:
        text = 'There are a few events coming up:'
    else:
        text = 'There are a bunch of events coming up:'
    if preamble and preamble.endswith(':') and 'events' in preamble.lower():
        text = preamble
    elif preamble:
        if preamble[-1] in ('.', '?', '!'):
            preamble += ' '
        else:
            preamble += ' \u2022 '
        text = text[0].lower() + text[1:]
        if 'event' in preamble.lower():
            text = 'Speaking of which, ' + text
        else:
            text = 'Also, ' + text
        text = preamble + text
    return '%s\n\n%s' % (text, '\n'.join(events))


def _handle_alerts(bot, records, groupid, alerts):
    # See https://github.com/nmlorg/metabot/issues/85.
    alerts = [
        alert for alert in alerts
        if alert['urgency'] in ('Immediate', 'Expected') and alert['severity'] in (
            'Extreme', 'Severe') and alert['certainty'] == 'Observed'
    ]

    text = _format_alerts(alerts)

    kwargs = {}
    key = ('alerts', groupid)
    if key in records:
        if not text:
            records.pop(key)
            return
        lasttext, lastmessage = records[key]
        if lasttext == text:
            return
        kwargs['reply_to_message_id'] = lastmessage['message_id']
    if text:
        message = None
        try:
            message = bot.send_message(chat_id=groupid,
                                       text=text,
                                       parse_mode='HTML',
                                       disable_web_page_preview=True,
                                       disable_notification=True,
                                       **kwargs)
        except ntelebot.errors.Error:
            logging.exception('While sending to %s:\n%s', groupid, text)
        if not message and kwargs:
            try:
                message = bot.send_message(chat_id=groupid,
                                           text=text,
                                           parse_mode='HTML',
                                           disable_web_page_preview=True,
                                           disable_notification=True)
            except ntelebot.errors.Error:
                logging.exception('While sending to %s:\n%s', groupid, text)
        if message:
            records[key] = (text, message)


def _format_alerts(alerts):
    if not alerts:
        return ''
    text = []
    for alert in alerts:
        text.append('<a href="https://alerts-v2.weather.gov/products/%s">%s %s</a>' %
                    (alert['id'], alert['event'], alert['id']))
        text.append('')
        desc = '%s\n\n%s\n\n%s' % (alert['description'], alert['instruction'], alert['areaDesc'])
        desc = re.sub(r'\s*\n\s*', ' ', re.sub(r'\s*\n\s*\n\s*', '\0',
                                               desc)).replace('\0', '\n\n').strip()
        if len(desc) > 250:
            desc = desc[:249] + '\u2026'
        text.append(desc)
        text.append('')
        text.append('')
    return '\n'.join(text).strip()



"""An ntelebot.loop.Loop that manages multiple bots."""

import logging

import ntelebot

from metabot import botconf
from metabot import datasettings
from metabot.calendars import multicalendar
from metabot.util import jsonutil
from metabot.util import msgbuilder


class MultiBot:
    """An ntelebot.loop.Loop that manages multiple bots."""

    def __init__(self, modules, confdir=None):
        self.dispatcher = _MultiBotLoopDispatcher()
        self.loop = ntelebot.loop.Loop()
        self.conf = botconf.BotConf(confdir)
        self.conf.finalize()
        self.multical = multicalendar.MultiCalendar()
        self.calendars = {}
        if confdir:
            for calendar_info in jsonutil.load(confdir + '/calendars.json') or ():
                cal = self.multical.add(calendar_info['calid'])
                self.calendars[cal.calcode] = calendar_info
        self.modules = {}
        for module in modules:
            modname = module.__name__.rsplit('.', 1)[-1]
            self.modules[modname] = module
            modinit = getattr(module, 'modinit', None)
            if modinit:
                modinit(self)

        for username, bot_config in self.conf['bots'].items():
            if bot_config['issue37']['telegram']['running']:
                self.run_bot(username)

    def add_bot(self, token):
        """Begin polling bot_config.token, dispatching updates through bot_config.modules."""

        username = ntelebot.bot.Bot(token).username
        self.conf['bots'][username] = {
            'issue37': {
                'telegram': {
                    'running': False,
                    'token': token,
                },
            },
        }
        self.conf.save()
        return username

    def _build_bot(self, username):
        bot_config = self.conf['bots'][username]
        bot = ntelebot.bot.Bot(bot_config['issue37']['telegram']['token'])
        bot.config = bot_config
        bot.multibot = self
        assert bot.username == username
        return bot

    def run_bot(self, username):
        """Begin polling for updates for the previously configured bot."""

        bot = self._build_bot(username)
        self.loop.add(bot, self.dispatcher)
        bot.config['issue37']['telegram']['running'] = True
        self.conf.save()

    def stop_bot(self, username):
        """Stop polling for updates for the referenced bot."""

        bot_config = self.conf['bots'][username]
        self.loop.remove(bot_config['issue37']['telegram']['token'])
        bot_config['issue37']['telegram']['running'] = False
        self.conf.save()

    def run(self):
        """Begin waiting for and dispatching updates sent to any bot currently running."""

        return self.loop.run()

    def stop(self):
        """Stop waiting for and dispatching updates sent to any bot currently running."""

        return self.loop.stop()


class _MultiBotLoopDispatcher(ntelebot.dispatch.LoopDispatcher):

    def __call__(self, bot, update):  # pylint: disable=too-many-branches,too-many-locals
        logging.info('%s', _pretty_repr(update))

        multibot = bot.multibot
        ctx = self.preprocessor(bot, update)
        if not ctx:
            return False

        with multibot.conf.record_mutations(ctx):
            msg = msgbuilder.MessageBuilder()

            ctx.botinfo = datasettings.DataSettings(bot.config['data'], bot.config['settings'])

            if not ctx.user:
                ctx.userinfo = None
            else:
                ctx.userinfo = datasettings.DataSettings(multibot.conf['users'][ctx.user['id']],
                                                         bot.config['users'][ctx.user['id']])
                for k, val in ctx.user.items():
                    if k not in ('first_name', 'id', 'last_name'):
                        ctx.userinfo.data[k] = val
                    ctx.userinfo.data.name = (
                        '%s %s' %
                        (ctx.user.get('first_name', ''), ctx.user.get('last_name', ''))).strip()

            if not ctx.chat or ctx.chat['type'] not in ('channel', 'group', 'supergroup'):
                ctx.groupinfo = None
            else:
                ctx.groupinfo = datasettings.DataSettings(multibot.conf['groups'][ctx.chat['id']],
                                                          bot.config['groups'][ctx.chat['id']])
                for k, val in ctx.chat.items():
                    if k != 'id':
                        ctx.groupinfo.data[k] = val

                if ctx.type == 'pin':
                    ctx.groupinfo.data.pinned_message_id = ctx.data['message_id']

            for modname, module in multibot.modules.items():
                modpredispatch = getattr(module, 'modpredispatch', None)
                if modpredispatch:
                    modpredispatch(ctx, msg, bot.config['issue37'][modname])

            ret = False
            for modname, module in multibot.modules.items():
                moddispatch = getattr(module, 'moddispatch', None)
                if moddispatch:
                    ret = moddispatch(ctx, msg, bot.config['issue37'][modname])
                    if ret is not False:
                        break

            if msg:
                msg.reply(ctx)

            return ret


def _pretty_repr(obj):
    if isinstance(obj, dict):
        return '{%s}' % ', '.join(
            '\033[31m%s\033[0m=%s' % (k, _pretty_repr(v)) for k, v in sorted(obj.items()))
    if isinstance(obj, (list, tuple)):
        return '[%s]' % ', '.join(map(_pretty_repr, obj))
    return '\033[32;1m%s\033[0m' % repr(obj)



"""Telegram-specific HTML sanitizer."""

import html
import html.parser
import logging


def cgi_escape(text):  # pylint: disable=missing-docstring
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>',
                                                                   '&gt;').replace('"', '&quot;')


class _HTMLSanitizer(html.parser.HTMLParser):
    __pieces = __strip = __current = None

    def sanitize(self, text, strip=False):
        """Sanitize (or optionally entirely strip) HTML for use in Telegram."""

        self.__pieces = []
        self.__strip = strip
        self.feed(text)
        self.close()
        return ''.join(self.__pieces)

    def handle_starttag(self, tag, attrs):
        if tag in ('br', 'div', 'p'):
            self.__pieces.append('\n')
            return
        if self.__strip or tag not in ('a', 'b', 'code', 'em', 'i', 'pre', 'strong'):
            return
        if self.__current:
            self.__pieces.append('</%s>' % self.__current)
        self.__current = tag
        if tag == 'a':
            attrs = dict(attrs)
            if attrs.get('href'):
                self.__pieces.append('<a href="%s">' % cgi_escape(attrs['href']))
            else:
                self.__pieces.append('<a>')
        else:
            self.__pieces.append('<%s>' % tag)

    def handle_endtag(self, tag):
        if tag == self.__current:
            self.__pieces.append('</%s>' % tag)
            self.__current = None

    def handle_data(self, data):
        self.__pieces.append(cgi_escape(data))

    def handle_entityref(self, name):  # pragma: no cover
        if name in ('lt', 'gt', 'amp', 'quot'):
            self.__pieces.append('&%s;' % name)
        else:
            codepoint = html.entities.name2codepoint.get(name)
            if codepoint:
                self.__pieces.append(chr(codepoint))
            else:
                self.__pieces.append('&amp;%s;' % name)

    def handle_charref(self, name):  # pragma: no cover
        name = name.lower()
        try:
            if name.startswith('x'):
                codepoint = int(name[1:], 16)
            else:
                codepoint = int(name)
            self.__pieces.append(cgi_escape(chr(codepoint)))
        except (OverflowError, ValueError):
            self.__pieces.append('&amp;#%s;' % name)

    def error(self, message):  # pragma: no cover  pylint: disable=missing-function-docstring
        # Remove when support for Python below 3.10 is dropped.
        logging.error('HTML sanitizer error: %s', message)


def sanitize(text, strip=False):
    """Sanitize (or optionally entirely strip) HTML for use in Telegram."""

    return _HTMLSanitizer().sanitize(text, strip=strip)



"""Find other group chats."""

import collections
import hashlib
import html
import re

import ntelebot
import requests


def cgi_escape(text):  # pylint: disable=missing-docstring
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>',
                                                                   '&gt;').replace('"', '&quot;')


ALIASES = ('channel', 'channels', 'group', 'groups', 'room', 'rooms')


def modhelp(unused_ctx, unused_modconf, sections):  # pylint: disable=missing-docstring
    sections['commands'].add('/groups \u2013 Find other group chats')


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if ctx.type in ('message', 'callback_query') and ctx.command in ALIASES:
        return default(ctx, msg, modconf)

    if ctx.type == 'inline_query' and ctx.prefix.lstrip('/') in ALIASES:
        return inline(ctx, modconf)

    return False


def default(ctx, msg, modconf):
    """Handle /groups."""

    ctx.private = True
    msg.path('/groups', 'Group List')

    groups_by_location = collections.defaultdict(list)
    for group in modconf['groups'].values():
        if group['invite_link'] or group['username']:
            groups_by_location[group['location'] or 'Worldwide'].append(group)

    if not groups_by_location:
        return msg.add("I don't know about any public groups yet, sorry!")

    location = ctx.text
    if location not in groups_by_location:
        msg.action = 'Choose a location'
        for location in sorted(groups_by_location, key=lambda name: tuple(reversed(name.split()))):
            msg.button(location, '/groups ' + location)
        return

    msg.path(location)

    for group in sorted(groups_by_location[location], key=lambda group: group['name']):
        invite_link = group['invite_link'] or 'https://t.me/' + group['username']
        message = '<a href="%s">%s</a>' % (invite_link, cgi_escape(group['name']))
        if group['desc']:
            message = '%s\n%s' % (message,
                                  re.sub('\\s*\n\\s*', ' \u2022 ', cgi_escape(group['desc'])))
        msg.add(message)


def inline(ctx, modconf):
    """Handle @BOTNAME groups."""

    terms = ctx.text.lower().split()[1:]
    results = []
    for group in sorted(modconf['groups'].values(), key=lambda group: group['name']):
        if len(results) >= 25:
            break
        if group['invite_link'] or group['username']:
            for term in terms:
                if term not in group['name'].lower() and term not in group['desc'].lower():
                    break
            else:
                invite_link = group['invite_link'] or 'https://t.me/' + group['username']
                message = '<a href="%s">%s</a>' % (invite_link, cgi_escape(group['name']))
                if group['desc']:
                    message = '%s\n%s' % (message, group['desc'])
                    title = '%s \u2022 %s' % (group['name'], group['username'] or
                                              group['invite_link'])
                    desc = group['desc']
                else:
                    title = group['name']
                    desc = invite_link
                results.append({
                    'type': 'article',
                    'id': invite_link,
                    'title': title,
                    'input_message_content': {
                        'message_text': message,
                        'parse_mode': 'HTML',
                        'disable_web_page_preview': True,
                    },
                    'description': desc,
                })
    ctx.reply_inline(results, cache_time=60)


def admin(frame):
    """Handle /admin BOTNAME groups."""

    ctx, msg, modconf = frame.ctx, frame.msg, frame.value
    text = frame.text
    if text.startswith('remove '):
        text = text[len('remove '):]
        group = modconf['groups'].pop(text)
        if group:
            msg.add('Removed <b>%s</b>.', group['name'])
        else:
            msg.add('Oops, the groups list changed since you loaded that screen, try again.')
    elif text:
        newgroup = get_group(ctx.bot, text)
        if not newgroup:
            msg.add(
                "I'm not sure what <code>%s</code> is! If it's a public group, it should have a "
                '<code>@USERNAME</code> code or <code>https://t.me/USERNAME</code> link in the '
                "group description screen. If it's a private group, and you are an admin, there "
                'should be a <code>https://t.me/joinchat/INVITECODE</code> link in the group '
                "description screen, otherwise you'll need to dig around or ask someone for it. "
                'Once you find it, paste it here:', text)
        else:
            for key, group in modconf['groups'].items():
                if (group['username'] and newgroup['username'] == group['username'] or
                        group['invite_link'] and newgroup['invite_link'] == group['invite_link']):
                    modconf['groups'][key] = newgroup
                    msg.add('Updated <b>%s</b>.', newgroup['name'])
                    break
            else:
                tmp = hashlib.sha1(newgroup['name'].encode('utf-8'))
                while tmp.hexdigest()[:6] in modconf['groups']:
                    tmp.update(b'.')
                modconf['groups'][tmp.hexdigest()[:6]] = newgroup
                msg.add('Added <b>%s</b>.', newgroup['name'])

    msg.action = 'Type a group username or invite link'
    msg.add('Type the <code>@USERNAME</code> or <code>https://t.me/joinchat/INVITE_LINK</code> for '
            'the group you want to add, or select an existing group to remove:')
    for key, group in sorted(modconf['groups'].items(), key=lambda ent: ent[1]['name']):
        msg.button('%s \u2022 %s' % (group['name'], group['username'] or group['invite_link']),
                   'remove %s' % key)


def get_group(bot, identifier):
    """Try to convert an @username or invite link to a normalized group info dict."""

    if identifier.startswith('tg://join?invite='):
        invite_link = 'https://t.me/joinchat/' + identifier[len('tg://join?invite='):]
        return get_private_group(invite_link)
    if identifier.startswith('https://t.me/joinchat/'):
        return get_private_group(identifier)
    if identifier.startswith('https://t.me/'):
        username = identifier[len('https://t.me/'):]
        return get_public_group(bot, username)
    return get_public_group(bot, identifier.lstrip('@'))


def get_public_group(bot, username):
    """Try to convert a username to a normalized group info dict."""

    try:
        info = bot.get_chat(chat_id='@' + username)
    except ntelebot.errors.Error:
        return

    group = {
        'desc': info.get('description', ''),
        'id': info['id'],
        'invite_link': '',
        'location': '',
        'name': info['title'],
        'type': info['type'],
        'username': info['username'],
    }

    return group


def get_private_group(invite_link):
    """Try to convert an invite link to a normalized group info dict."""

    info = fetch_opengraph(invite_link)
    if not info.get('title'):
        return

    return {
        'desc': info['description'],
        'id': '',
        'invite_link': invite_link,
        'location': '',
        'name': info['title'],
        'type': '',
        'username': '',
    }


def fetch_opengraph(url):
    """Basic (and fragile) Open Graph object fetcher/decoder."""

    values = {}
    for line in requests.get(url, timeout=10).text.splitlines():
        ret = re.search('<meta property="og:([^"]+)" content="([^"]*)">', line)
        if ret:
            key, value = ret.groups()
            values[html.unescape(key)] = html.unescape(value)
    return values



"""Display recent and upcoming events."""

import pytz

from metabot.util import adminui
from metabot.util import eventutil
from metabot.util import html
from metabot.util import humanize

ALIASES = ('calendar', 'event', 'events')


def modhelp(unused_ctx, unused_modconf, sections):  # pylint: disable=missing-docstring
    sections['commands'].add('/events \u2013 Display recent and upcoming events')


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if ctx.type in ('message', 'callback_query') and ctx.command in ALIASES:
        if ctx.chat['type'] != 'private':
            return group(ctx, msg)
        if ctx.prefix == 'set':
            return settings(ctx, msg, modconf)
        if ctx.prefix == 'admin':
            return customize(ctx, msg, modconf)
        return private(ctx, msg, modconf)

    if ctx.type == 'inline_query' and ctx.prefix.lstrip('/') in ALIASES:
        return inline(ctx, modconf)

    return False


def group(ctx, msg):
    """Handle /events in a group chat."""

    group_id = '%s' % ctx.chat['id']
    groupconf = ctx.bot.config['issue37']['moderator'][group_id]
    calcodes, tzinfo, count, days, unused_hour, unused_dow = eventutil.get_group_conf(groupconf)
    if not calcodes or not tzinfo:
        missing = []
        if not calcodes:
            missing.append('choose one or more calendars')
        if not tzinfo:
            missing.append('set the time zone')
        return msg.add(
            "I'm not configured for this group! Ask a bot admin to go into the <b>moderator</b> "
            'module settings, group <b>%s</b>, and %s.', group_id, humanize.list(missing))

    events, unused_alerts = eventutil.get_group_events(ctx.bot, calcodes, tzinfo, count, days)
    if not events:
        msg.add('No events in the next %s days!', days)
    else:
        url = eventutil.get_image(events[0], ctx.bot.config)
        if url:
            msg.add('photo:' + url)
        msg.add('\n'.join(
            eventutil.format_event(ctx.bot, event, tzinfo, full=False) for event in events))


def private(ctx, msg, modconf):  # pylint: disable=too-many-branches,too-many-locals
    """Handle /events in a private chat."""

    eventid, timezone = ctx.split(2)
    if ':' in eventid and timezone:
        suffix = ' ' + timezone
        calcodes = eventid.split(':', 1)[0]
    else:
        suffix = ''
        user_id = '%s' % ctx.user['id']
        userconf = modconf['users'][user_id]
        calcodes = userconf.get('calendars')
        timezone = userconf.get('timezone')
        if not calcodes or not timezone:
            missing = []
            if not calcodes:
                missing.append('choose one or more calendars')
            if not timezone:
                missing.append('set your time zone')
            msg.add('Please %s!', humanize.list(missing))
            return msg.button('Settings', '/events set')

    calendar_view = ctx.bot.multibot.multical.view(calcodes.split())
    tzinfo = pytz.timezone(timezone)

    prevev, event, nextev = calendar_view.get_event(eventid)
    if not event:
        prevev, event, nextev = calendar_view.get_event()
    if not event:
        msg.add('No upcoming events!')
    else:
        msg.add(eventutil.format_event(ctx.bot, event, tzinfo, full=True))
    if ctx.user['id'] in ctx.bot.config['issue37']['admin']['admins']:
        msg.button('Customize', '/events admin ' + event['local_id'])
    buttons = [None, ('Settings', '/events set'), None]
    if prevev:
        buttons[0] = ('Prev', '/events %s%s' % (prevev['local_id'], suffix))
    if suffix:
        buttons[1] = ('My Events', '/events')
    elif event and event['local_id'] != calendar_view.current_local_id:
        buttons[1] = ('Current', '/events')
    if nextev:
        buttons[2] = ('Next', '/events %s%s' % (nextev['local_id'], suffix))
    msg.buttons(buttons)


def inline(ctx, modconf):  # pylint: disable=too-many-branches,too-many-locals
    """Handle @BOTNAME events."""

    user_id = '%s' % ctx.user['id']
    userconf = modconf['users'][user_id]
    calcodes = userconf.get('calendars')
    timezone = userconf.get('timezone')
    if not calcodes or not timezone:
        missing = []
        if not calcodes:
            missing.append('choose one or more calendars')
        if not timezone:
            missing.append('set your time zone')
        return ctx.reply_inline([],
                                is_personal=True,
                                cache_time=30,
                                switch_pm_text='Click to %s!' % humanize.list(missing),
                                switch_pm_parameter='L2V2ZW50cw')

    calendar_view = ctx.bot.multibot.multical.view(calcodes.split())
    tzinfo = pytz.timezone(timezone)

    terms = ctx.text.lower().split()[1:]
    full = False
    if terms and terms[0].lower() == 'full':
        terms.pop(0)
        full = True
    nextid = None
    results = []
    while len(results) < 25:
        _, event, nextev = calendar_view.get_event(nextid)
        nextid = nextev and nextev['local_id']
        if not event:
            break
        if full:
            text = ('%s %s' % (event['summary'], event['description'])).lower()
        else:
            text = event['summary'].lower()
        for term in terms:
            if term not in text:
                break
        else:
            subtitle = eventutil.humanize_range(event['start'], event['end'], tzinfo)
            if event['location']:
                subtitle = '%s @ %s' % (subtitle, event['location'].split(',', 1)[0])
            if full and event['description']:
                title = '%s \u2022 %s' % (event['summary'], subtitle)
                description = html.sanitize(event['description'], strip=True)
            else:
                title = event['summary']
                description = subtitle
            results.append({
                'description': description,
                'input_message_content': {
                    'disable_web_page_preview': True,
                    'message_text': eventutil.format_event(ctx.bot, event, tzinfo, full=full),
                    'parse_mode': 'HTML',
                },
                'id': event['local_id'],
                #'thumb_url': icon,
                'title': title,
                'type': 'article',
            })
        if not nextid:
            break
    ctx.reply_inline(results,
                     is_personal=True,
                     cache_time=30,
                     switch_pm_text='Settings',
                     switch_pm_parameter='L2V2ZW50cyBzZXQ')


def customize(ctx, msg, modconf):
    """Handle /events admin."""

    _, eventid, field, text = ctx.split(4)

    msg.path('/events', 'Events')
    msg.path('admin', 'Customize')
    msg.path(eventid)

    _, event, _ = ctx.bot.multibot.multical.get_event(eventid)
    if not event:
        return msg.add('Unknown event!')

    frame = adminui.Frame(ctx, msg, modconf, 'events', None, field)
    menu = adminui.Menu(
        ('event', 'event', 'What banner image should be used for this specific event?'),
        ('series', 'series', 'What banner image should be used for events with this title?'),
    )
    _, handler = menu.select(frame)
    if handler == 'event':
        msg.path('event')
        frame = adminui.Frame(ctx, msg, modconf['events'], eventid, None, text)
        adminui.image(frame)
    elif handler == 'series':
        msg.path('series')
        frame = adminui.Frame(ctx, msg, modconf['series'], event['summary'], None, text)
        adminui.image(frame)
    else:
        menu.display(frame)


def settings(ctx, msg, modconf):
    """Handle /events set."""

    _, text = ctx.split(2)

    msg.path('/events', 'Events')
    msg.path('set', 'Settings')

    user_id = '%s' % ctx.user['id']
    adminui.Menu(
        ('calendars', adminui.calendars, 'Which calendars do you want to see?'),
        ('timezone', adminui.timezone, 'What time zone are you in?'),
    ).handle(adminui.Frame(ctx, msg, modconf['users'], user_id, None, text))



"""Tests for metabot.modules.reminders."""

import pytest

from metabot.calendars import loader
from metabot.modules import events
from metabot.modules import moderator
from metabot.modules import reminders


@pytest.fixture
def conversation(build_conversation, monkeypatch):  # pylint: disable=missing-docstring
    monkeypatch.setattr('time.time', lambda: 2000.)

    cal = loader.get('static:test_events')
    assert cal.calcode == '6fc2c510'
    cal.events = {
        '6fc2c510:alpha': {
            'description': 'Alpha Description',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        },
        '6fc2c510:bravo': {
            'description': 'Bravo Description',
            'end': 60 * 60 * 24 * 7 + 60 * 60,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 60 * 60 * 24 * 7,
            'summary': 'Bravo Summary',
        },
        '6fc2c510:charlie': {
            'description': 'Charlie Description',
            'end': 60 * 60 * 24 * 7 + 1 + 60 * 60,
            'local_id': '6fc2c510:charlie',
            'location': 'Charlie Venue, Rest of Charlie Location',
            'start': 60 * 60 * 24 * 7 + 1,
            'summary': 'Charlie Summary',
        },
    }

    conv = build_conversation(events, moderator)
    conv.multibot.multical.add('static:test_events')
    conv.multibot.calendars['6fc2c510'] = {'name': 'Test Calendar'}
    return conv


# pylint: disable=line-too-long


def test_format_daily_message():  # pylint: disable=missing-docstring
    # yapf: disable - pylint: disable=protected-access
    assert reminders._format_daily_message('', ['EVENT1']) == (
        "There's an event coming up:\n"
        '\n'
        'EVENT1')
    assert reminders._format_daily_message('', ['EVENT1', 'EVENT2']) == (
        "There are a couple events coming up:\n"
        '\n'
        'EVENT1\n'
        'EVENT2')
    assert reminders._format_daily_message('', ['EVENT1', 'EVENT2', 'EVENT3']) == (
        "There are a few events coming up:\n"
        '\n'
        'EVENT1\n'
        'EVENT2\n'
        'EVENT3')
    assert reminders._format_daily_message('', ['EVENT1', 'EVENT2', 'EVENT3', 'EVENT4']) == (
        "There are a bunch of events coming up:\n"
        '\n'
        'EVENT1\n'
        'EVENT2\n'
        'EVENT3\n'
        'EVENT4')
    assert reminders._format_daily_message('', ['EVENT1', 'EVENT2', 'EVENT3', 'EVENT4', 'EVENT5']) == (
        "There are a bunch of events coming up:\n"
        '\n'
        'EVENT1\n'
        'EVENT2\n'
        'EVENT3\n'
        'EVENT4\n'
        'EVENT5')

    assert reminders._format_daily_message('1 + 1 = 2!', ['EVENT1']) == (
        "1 + 1 = 2! Also, there's an event coming up:\n"
        '\n'
        'EVENT1')

    assert reminders._format_daily_message('Visit @MYGROUP', ['EVENT1']) == (
        "Visit @MYGROUP • Also, there's an event coming up:\n"
        '\n'
        'EVENT1')

    assert reminders._format_daily_message('I love events.', ['EVENT1']) == (
        "I love events. Speaking of which, there's an event coming up:\n"
        '\n'
        'EVENT1')

    assert reminders._format_daily_message('I love events:', ['EVENT1']) == (
        "I love events:\n"
        '\n'
        'EVENT1')
    # yapf: enable


@pytest.fixture
def daily_messages(conversation, monkeypatch):  # pylint: disable=missing-docstring,redefined-outer-name
    monkeypatch.setattr('time.time', lambda: 0)

    replies = conversation.raw_message('/dummy')
    assert conversation.format_messages(replies) == ''

    records = {}

    def _daily_messages(initial=False):
        replies.clear()
        if initial:
            records.clear()
            conversation.bot.config['issue37']['moderator']['-1002000002000']['daily']['hour'] = 0
        else:
            conversation.bot.config['issue37']['moderator']['-1002000002000']['daily']['hour'] = 1
        reminders._daily_messages(conversation.multibot, records)  # pylint: disable=protected-access
        return records, conversation.format_messages(replies)

    assert _daily_messages(True) == ({}, '')

    conversation.bot.config['issue37']['moderator']['-1002000002000'] = {
        'calendars': '6fc2c510',
        'daily': {
            'hour': 0,
        },
        'timezone': 'UTC',
    }

    return _daily_messages


def test_daily_messages(daily_messages):  # pylint: disable=redefined-outer-name
    """Test basic daily announcement logic (essentially, test the daily_messages fixture)."""

    assert daily_messages(True) == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Alpha Description',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        }, {
            'description': 'Bravo Description',
            'end': 608400,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 604800,
            'summary': 'Bravo Summary',
        }], {
            'message_id': 12345,
        }),
    }, """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
""")

    assert daily_messages() == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Alpha Description',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        }, {
            'description': 'Bravo Description',
            'end': 608400,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 604800,
            'summary': 'Bravo Summary',
        }], {
            'message_id': 12345,
        }),
    }, '')


def test_daily_messages_updated(daily_messages):  # pylint: disable=redefined-outer-name
    """Test daily announcement update messages."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['summary'] = 'Edited Summary'
    cal.events['6fc2c510:alpha']['end'] += 60

    assert daily_messages() == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Alpha Description',
            'end': 2060,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Edited Summary',
        }, {
            'description': 'Bravo Description',
            'end': 608400,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 604800,
            'summary': 'Bravo Summary',
        }], {
            'message_id': 12345,
        }),
    }, """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • Edited Summary
      <s>Alpha Summary</s>
      Edited Summary
      <s>…2:16–12:33ᵃᵐ</s>
      …2:16–12:34ᵃᵐ


[chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Edited Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:34ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12ᵃᵐ</a>]
""")

    assert daily_messages() == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Alpha Description',
            'end': 2060,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Edited Summary',
        }, {
            'description': 'Bravo Description',
            'end': 608400,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 604800,
            'summary': 'Bravo Summary',
        }], {
            'message_id': 12345,
        }),
    }, '')


def test_daily_messages_future(daily_messages, monkeypatch):  # pylint: disable=redefined-outer-name
    """Test the behavior when an event transitions from TODAY to NOW."""

    daily_messages(True)

    monkeypatch.setattr('time.time', lambda: 1000)

    assert daily_messages()[1] == ''

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['summary'] = 'Now Summary'

    # Note that TODAY is displayed as NOW now:
    assert daily_messages()[1] == """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • Now Summary
      <s>Alpha Summary</s>
      Now Summary


[chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Now Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">NOW, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12:16ᵃᵐ</a>]
"""


def test_daily_messages_multiline(daily_messages):  # pylint: disable=redefined-outer-name
    """Verify multiline event components are rendered in updates concisely."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = 'Multi\n\nLine\nDescription'

    assert daily_messages() == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Multi\n\nLine\nDescription',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        }, {
            'description': 'Bravo Description',
            'end': 608400,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 604800,
            'summary': 'Bravo Summary',
        }], {
            'message_id': 12345,
        }),
    }, """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • Alpha Summary
      <s>Alpha Description</s>
      Multi Line Description


[chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12ᵃᵐ</a>]
""")


def test_daily_messages_add_remove_event(conversation, daily_messages):  # pylint: disable=redefined-outer-name
    """Verify what happens when events are added/removed entirely."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:new'] = {
        'description': 'New Description',
        'end': 3000,
        'local_id': '6fc2c510:new',
        'location': 'New Venue, Rest of New Location',
        'start': 2000,
        'summary': 'New Summary',
    }
    conversation.multibot.multical.ordered.append(cal.events['6fc2c510:new'])
    cal.events.pop('6fc2c510:bravo')
    assert conversation.multibot.multical.ordered[1]['local_id'] == '6fc2c510:bravo'
    assert conversation.multibot.multical.ordered.pop(1)
    conversation.multibot.multical._rebuild()  # pylint: disable=protected-access

    assert daily_messages() == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Alpha Description',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        }, {
            'description': 'New Description',
            'end': 3000,
            'local_id': '6fc2c510:new',
            'location': 'New Venue, Rest of New Location',
            'start': 2000,
            'summary': 'New Summary',
        }], {
            'message_id': 12345,
        }),
    }, """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • New Summary
    ◦ New event!
  • Bravo Summary
    ◦ Removed.


[chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>New Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpuZXcgVVRD">TODAY, Thu 1ˢᵗ, 12:33–12:50ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Venue%2C+Rest+of+New+Location">New Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12ᵃᵐ</a>]
""")


def test_daily_messages_ignored(daily_messages):  # pylint: disable=redefined-outer-name
    """Verify non-notification-worthy changes are properly ignored."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = (
        'Alpha Description https://www.example.com/?id=1234')

    assert daily_messages()[1] == ''

    cal.events['6fc2c510:alpha']['updated'] = 23456

    assert daily_messages()[1] == ''


def test_daily_messages_icons(conversation, daily_messages):  # pylint: disable=redefined-outer-name
    """Run through all the cases where events trigger banner images."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = 'Board Games!'

    # An update does not include references to icons.
    assert daily_messages()[1] == """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • Alpha Summary
      <s>Alpha Description</s>
      Board Games!


[chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12ᵃᵐ</a>]
"""

    # But an initial announcement does (photo=...).
    assert daily_messages(True)[1] == """\
[chat_id=-1002000002000 disable_notification=True parse_mode=HTML photo=https://ssl.gstatic.com/calendar/images/eventillustrations/v1/img_gamenight_2x.jpg]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    conversation.bot.config['issue37']['events']['series']['ha sum'] = 'SERIES ICON'

    assert daily_messages(True)[1] == """\
[chat_id=-1002000002000 disable_notification=True parse_mode=HTML photo=SERIES ICON]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    conversation.bot.config['issue37']['events']['events']['6fc2c510:alpha'] = 'EVENT ICON'

    assert daily_messages(True)[1] == """\
[chat_id=-1002000002000 disable_notification=True parse_mode=HTML photo=EVENT ICON]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    cal.events['6fc2c510:alpha']['description'] = 'Fun Games!'

    # Removing an icon trigger leaves the image in place.
    assert daily_messages()[1] == """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • Alpha Summary
      <s>Board Games!</s>
      Fun Games!


[chat_id=-1002000002000 message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12ᵃᵐ</a>]
"""


def test_daily_messages_geometry(conversation, daily_messages):  # pylint: disable=redefined-outer-name
    """See https://github.com/nmlorg/metabot/issues/75."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = 'Trigger'
    conversation.bot.config['issue37']['moderator']['-1002000002000']['maxeventscount'] = 1

    assert daily_messages() == ({
        ('modulestestbot', '-1002000002000'): (0, [{
            'description': 'Trigger',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        }], {
            'message_id': 12345,
        }),
    }, """\
[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
  • Alpha Summary
      <s>Alpha Description</s>
      Trigger


[chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There's an event coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">TODAY, Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>

[<a href="https://t.me/c/2000002000/12345">Updated 12ᵃᵐ</a>]
""")


def test_quick_diff():
    """Test description string differ."""

    # pylint: disable=protected-access
    assert reminders._quick_diff('', '') is None
    assert reminders._quick_diff('0987654321', '0987654321new') == ('0987654321', '0987654321new')
    assert reminders._quick_diff('10987654321', '10987654321new') == ('…987654321', '…987654321new')
    assert reminders._quick_diff('10987654321old', '10987654321') == ('…987654321old', '…987654321')
    assert reminders._quick_diff('10987654321old',
                                 '10987654321new') == ('…987654321old', '…987654321new')
    assert reminders._quick_diff(
        'x',
        'x234567890123456789012345678901234567890') == ('x',
                                                        'x234567890123456789012345678901234567890')
    assert reminders._quick_diff(
        'x',
        'x2345678901234567890123456789012345678901') == ('x',
                                                         'x23456789012345678901234567890123456789…')


@pytest.fixture
def handle_alerts(conversation):  # pylint: disable=missing-docstring,redefined-outer-name
    replies = conversation.raw_message('/dummy')
    assert conversation.format_messages(replies) == ''

    records = {}
    groupid = '-1001000001000'

    def _handle_alerts(alerts):
        replies.clear()
        reminders._handle_alerts(conversation.bot, records, groupid, alerts)  # pylint: disable=protected-access
        return conversation.format_messages(replies)

    return _handle_alerts


def test_handle_alerts(handle_alerts):  # pylint: disable=redefined-outer-name
    """Test basic behavior of reminders._handle_alerts."""

    # First poll; no active alerts.
    assert handle_alerts([]) == ''

    # Second poll; brand-new alert.
    alerts = [{
        'id': 'alert-id',
        'areaDesc': 'Area description',
        'description': 'Alert Description',
        'instruction': 'Alert Instructions',
        'event': 'Winter Storm Warning',
        'messageType': 'Alert',
        'urgency': 'Immediate',
        'severity': 'Extreme',
        'certainty': 'Observed',
    }]
    assert handle_alerts(alerts) == """\
[chat_id=-1001000001000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
<a href="https://alerts-v2.weather.gov/products/alert-id">Winter Storm Warning alert-id</a>

Alert Description

Alert Instructions

Area description
"""

    # Third poll; same alert is active.
    assert handle_alerts(alerts) == ''

    # Fourth poll; existing alert is modified.
    alerts[0]['instruction'] = 'Run for your lives!'
    assert handle_alerts(alerts) == """\
[chat_id=-1001000001000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
<a href="https://alerts-v2.weather.gov/products/alert-id">Winter Storm Warning alert-id</a>

Alert Description

Run for your lives!

Area description
"""

    # Fifth poll; another new alert.
    alerts.append({
        'id': 'alert-id-2',
        'areaDesc': 'Second area description',
        'description': 'Second Description',
        'instruction': 'Second Instructions',
        'event': 'Winter Storm Watch',
        'messageType': 'Alert',
        'urgency': 'Immediate',
        'severity': 'Extreme',
        'certainty': 'Observed',
    })
    assert handle_alerts(alerts) == """\
[chat_id=-1001000001000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
<a href="https://alerts-v2.weather.gov/products/alert-id">Winter Storm Warning alert-id</a>

Alert Description

Run for your lives!

Area description


<a href="https://alerts-v2.weather.gov/products/alert-id-2">Winter Storm Watch alert-id-2</a>

Second Description

Second Instructions

Second area description
"""

    # Sixth poll; both alerts have been canceled.
    assert handle_alerts([]) == ''

    # Seventh poll; a new alert that we don't care about.
    alerts = [{
        'id': 'alert-id-3',
        'areaDesc': 'Third area description',
        'description': 'Third Description',
        'instruction': 'Third Instructions',
        'event': 'Special Weather Statement',
        'messageType': 'Alert',
        'urgency': 'Immediate',
        'severity': 'Moderate',
        'certainty': 'Observed',
    }]
    assert handle_alerts([]) == ''



"""Tests for metabot.util.humanize."""

import datetime

from metabot.util import humanize


def test_humanize_date():
    """Quick tests for humanize.date."""

    base = datetime.date(2017, 11, 15)
    assert humanize.date(datetime.date(2017, 11, 15), base=base) == 'Wed 15ᵗʰ'
    assert humanize.date(datetime.date(2017, 11, 16), base=base) == 'Thu 16ᵗʰ'


def test_humanize_dayofmonth():
    """Quick tests for humanize.dayofmonth."""

    assert humanize.dayofmonth(1) == '1ˢᵗ'
    assert humanize.dayofmonth(2) == '2ⁿᵈ'
    assert humanize.dayofmonth(3) == '3ʳᵈ'
    assert humanize.dayofmonth(4) == '4ᵗʰ'
    assert humanize.dayofmonth(11) == '11ᵗʰ'
    assert humanize.dayofmonth(12) == '12ᵗʰ'
    assert humanize.dayofmonth(13) == '13ᵗʰ'
    assert humanize.dayofmonth(21) == '21ˢᵗ'


def test_nextmonth():
    """Quick tests for humanize._nextmonth."""

    # pylint: disable=protected-access
    assert humanize._nextmonth(datetime.date(2024, 11, 1)) == datetime.date(2024, 12, 1)
    assert humanize._nextmonth(datetime.date(2024, 12, 1), -1) == datetime.date(2024, 11, 1)
    assert humanize._nextmonth(datetime.date(2024, 12, 1)) == datetime.date(2025, 1, 1)
    assert humanize._nextmonth(datetime.date(2025, 1, 1), -1) == datetime.date(2024, 12, 1)

    assert humanize._nextmonth(datetime.date(2025, 1, 31), 1) == datetime.date(2025, 2, 28)
    assert humanize._nextmonth(datetime.date(2025, 1, 31), 2) == datetime.date(2025, 3, 31)
    assert humanize._nextmonth(datetime.date(2025, 1, 31), 3) == datetime.date(2025, 4, 30)
    assert humanize._nextmonth(datetime.date(2025, 1, 31), 4) == datetime.date(2025, 5, 31)

    assert humanize._nextmonth(datetime.date(2025, 5, 31), -1) == datetime.date(2025, 4, 30)
    assert humanize._nextmonth(datetime.date(2025, 5, 31), -2) == datetime.date(2025, 3, 31)
    assert humanize._nextmonth(datetime.date(2025, 5, 31), -3) == datetime.date(2025, 2, 28)
    assert humanize._nextmonth(datetime.date(2025, 5, 31), -4) == datetime.date(2025, 1, 31)


def test_humanize_howrecent():
    """Quick tests for humanize.howrecent."""

    base = datetime.datetime(2017, 11, 15, 12)
    assert humanize.howrecent(datetime.datetime(2017, 11, 15, 10),
                              datetime.datetime(2017, 11, 15, 11),
                              base=base) == 'TODAY,'
    assert humanize.howrecent(datetime.datetime(2017, 11, 15, 11),
                              datetime.datetime(2017, 11, 15, 12),
                              base=base) == 'NOW,'
    assert humanize.howrecent(datetime.datetime(2017, 11, 15, 12),
                              datetime.datetime(2017, 11, 15, 13),
                              base=base) == 'NOW,'
    assert humanize.howrecent(datetime.datetime(2017, 11, 15, 13),
                              datetime.datetime(2017, 11, 15, 14),
                              base=base) == 'TODAY,'

    yesterday = datetime.datetime(2017, 11, 14, 12)
    assert humanize.howrecent(datetime.datetime(2017, 11, 14), yesterday, base=base) == 'YESTERDAY,'
    assert humanize.howrecent(datetime.datetime(2017, 11, 13), yesterday, base=base) == '² ᵈᵃʸˢ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 9), yesterday, base=base) == '⁶ ᵈᵃʸˢ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 8), yesterday, base=base) == '¹ʷ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 2), yesterday, base=base) == '¹ʷ⁶ᵈ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 1), yesterday, base=base) == '²ʷ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2016, 11, 16), yesterday, base=base) == '¹¹ᵐ⁴ʷ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2016, 11, 15), yesterday, base=base) == '¹ʸ ᵃᵍᵒ'
    assert humanize.howrecent(datetime.datetime(2016, 10, 15), yesterday, base=base) == '¹ʸ¹ᵐ ᵃᵍᵒ'

    future = datetime.datetime(2018, 3, 30)
    assert humanize.howrecent(datetime.datetime(2017, 11, 16), future, base=base) == 'TOMORROW,'
    assert humanize.howrecent(datetime.datetime(2017, 11, 17), future, base=base) == '² ᵈᵃʸˢ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 21), future, base=base) == '⁶ ᵈᵃʸˢ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 22), future, base=base) == '¹ʷ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 28), future, base=base) == '¹ʷ⁶ᵈ'
    assert humanize.howrecent(datetime.datetime(2017, 11, 29), future, base=base) == '²ʷ'
    assert humanize.howrecent(datetime.datetime(2017, 12, 14), future, base=base) == '⁴ʷ¹ᵈ'
    assert humanize.howrecent(datetime.datetime(2017, 12, 15), future, base=base) == '¹ᵐ'
    assert humanize.howrecent(datetime.datetime(2017, 12, 16), future, base=base) == '¹ᵐ'
    assert humanize.howrecent(datetime.datetime(2017, 12, 22), future, base=base) == '¹ᵐ¹ʷ'
    assert humanize.howrecent(datetime.datetime(2018, 1, 15), future, base=base) == '²ᵐ'
    assert humanize.howrecent(datetime.datetime(2018, 1, 21), future, base=base) == '²ᵐ'
    assert humanize.howrecent(datetime.datetime(2018, 11, 14), future, base=base) == '¹¹ᵐ⁴ʷ'
    assert humanize.howrecent(datetime.datetime(2018, 11, 15), future, base=base) == '¹ʸ'
    assert humanize.howrecent(datetime.datetime(2018, 12, 14), future, base=base) == '¹ʸ'
    assert humanize.howrecent(datetime.datetime(2019, 12, 15), future, base=base) == '²ʸ¹ᵐ'


def test_humanize_list():
    """Quick tests for humanize.list."""

    assert humanize.list(['one']) == 'one'
    assert humanize.list(['one', 'two']) == 'one and two'
    assert humanize.list(['one', 'two', 'three']) == 'one, two, and three'


def test_humanize_range(monkeypatch):
    """Quick tests for humanize.range (and humanize.date/humanize.time)."""

    def _test(start, end):
        return humanize.range(start, end).replace('\u2013', '-')

    start = datetime.date(2017, 11, 15)
    start_ts = float(start.strftime('%s'))
    monkeypatch.setattr('time.time', lambda: start_ts)
    next_day = datetime.date(2017, 11, 16)
    next_week = datetime.date(2017, 11, 22)
    next_month = datetime.date(2017, 12, 15)
    next_year = datetime.date(2018, 1, 15)

    assert _test(start, start) == 'NOW, Wed 15ᵗʰ'
    assert _test(start, next_day) == 'NOW, Wed 15-16ᵗʰ'
    assert _test(start, next_week) == 'NOW, Wed 15-22ⁿᵈ'
    assert _test(start, next_month) == 'NOW, Wed 15ᵗʰ - Fri, Dec 15ᵗʰ'
    assert _test(start, next_year) == 'NOW, Wed 15ᵗʰ - Mon, Jan (2018) 15ᵗʰ'

    start = datetime.datetime(2017, 11, 15, 6)
    start_ts = float(start.strftime('%s'))
    monkeypatch.setattr('time.time', lambda: start_ts)
    next_min = datetime.datetime(2017, 11, 15, 6, 1)
    next_hour = datetime.datetime(2017, 11, 15, 7)
    next_pm = datetime.datetime(2017, 11, 15, 18)
    hours_23 = datetime.datetime(2017, 11, 16, 5)
    hours_25 = datetime.datetime(2017, 11, 16, 7)
    next_month = datetime.datetime(2017, 12, 15, 6)
    next_year = datetime.datetime(2018, 1, 15, 6)

    assert _test(start, next_min) == 'NOW, Wed 15ᵗʰ, 6-6:01ᵃᵐ'
    assert _test(start, next_hour) == 'NOW, Wed 15ᵗʰ, 6-7ᵃᵐ'
    assert _test(start, next_pm) == 'NOW, Wed 15ᵗʰ, 6ᵃᵐ - 6ᵖᵐ'
    assert _test(start, hours_23) == 'NOW, Wed 15ᵗʰ, 6ᵃᵐ - 5ᵃᵐ'
    assert _test(start, hours_25) == 'NOW, Wed 15ᵗʰ, 6ᵃᵐ - Thu 16ᵗʰ, 7ᵃᵐ'
    assert _test(start, next_month) == 'NOW, Wed 15ᵗʰ, 6ᵃᵐ - Fri, Dec 15ᵗʰ, 6ᵃᵐ'
    assert _test(start, next_year) == 'NOW, Wed 15ᵗʰ, 6ᵃᵐ - Mon, Jan (2018) 15ᵗʰ, 6ᵃᵐ'



"""Simplified interface to https://docs.python.org/2/library/pickle.html."""

import pickle
import pickletools


def load(fname):
    """Load fname as a Pickle file, silently returning None on any error."""

    try:
        data = open(fname, 'rb').read()
    except IOError:
        return

    try:
        return pickle.loads(data)
    except (EOFError, pickle.UnpicklingError):
        pass


def optimize(obj, store=None):
    """Return a copy of obj with normalized and deduplicated strings."""

    if store is None:
        store = {}

    opt = lambda obj: optimize(obj, store=store)

    if isinstance(obj, dict):
        new = [(opt(k), opt(v)) for k, v in obj.items()]
        obj.clear()
        obj.update(new)
        return obj
    if isinstance(obj, list):
        for i, value in enumerate(obj):
            obj[i] = opt(value)
        return obj
    if isinstance(obj, tuple):
        return tuple(map(opt, obj))
    if isinstance(obj, (bytes, str)):
        new = store.get(obj)
        if new is None:
            store[obj] = new = obj
        return new
    return obj


def dump(fname, obj, store=None):
    """Optimize obj, then save it as a Pickle file to fname."""

    obj = optimize(obj, store=store)
    data = pickletools.optimize(pickle.dumps(obj, -1))
    with open(fname, 'wb') as fobj:
        fobj.write(data)
    return obj



"""Base calendar container."""

import hashlib
import logging


class Calendar:
    """Base calendar container."""

    caltype = 'base'
    last_update = None

    def __init__(self, calid):
        caltype, self.calpath = calid.split(':', 1)
        assert caltype == self.caltype
        self.calcode = self._hashid(calid)
        self.events = {}

    @staticmethod
    def _hashid(uid):
        return hashlib.sha1(uid.encode('ascii')).hexdigest()[:8]

    def _normalize(self, local):
        location = local.get('location') or ''
        if location in ('TBA', 'TBD', '(none)'):
            location = ''
        elif location:
            location = location.replace('\n', ', ')
        local_id = local.get('local_id') or '%s:%s' % (self.calcode, self._hashid(local['id']))

        return {
            'description': local.get('description') or '',
            'end': local['end'],
            'id': local['id'],
            'local_id': local_id,
            'location': location,
            'start': local['start'],
            'summary': local.get('summary') or '',
            'updated': local['updated'],
        }

    def _removed(self, proto_id):
        local_id = '%s:%s' % (self.calcode, self._hashid(proto_id))
        local = self.events.pop(local_id, None)
        if local:
            logging.info('Event %s (%s) was removed.', local['id'], local['summary'])
            return True

    def _updated(self, proto):
        local = self._normalize(self.event_proto_to_local(proto))
        oldlocal = self.events.get(local['local_id'])
        if not oldlocal:
            logging.info('Event %s (%s) added.', local['id'], local['summary'])
        elif oldlocal != local:
            logging.info('Event %s (%s) updated.', local['id'], local['summary'])
        else:
            return
        self.events[local['local_id']] = local
        if not self.last_update or self.last_update < local['updated']:
            self.last_update = local['updated']
        return True

    def add(self, local):
        """Add an event (in local format)."""

        newlocal = self._normalize(
            self.event_proto_to_local(self.proto_add(self.event_local_to_proto(local))))
        logging.info('Added event %s (%s).', newlocal['id'], newlocal['summary'])
        self.events[newlocal['local_id']] = newlocal
        if not self.last_update or self.last_update < newlocal['updated']:
            self.last_update = newlocal['updated']
        return newlocal

    def remove(self, local_id):
        """Remove an event (given as calcode:localcode)."""

        currentlocal = self.events.pop(local_id)
        self.proto_remove(currentlocal['id'])
        logging.info('Removed event %s (%s).', currentlocal['id'], currentlocal['summary'])

    def update(self, local_id, local):
        """Update an event (given as calcode:localcode, in local format)."""

        currentlocal = self.events[local_id]
        newlocal = self._normalize(
            self.event_proto_to_local(
                self.proto_update(currentlocal['id'], self.event_local_to_proto(local))))
        logging.info('Updated event %s (%s).', newlocal['id'], newlocal['summary'])
        self.events[newlocal['local_id']] = newlocal
        if not self.last_update or self.last_update < newlocal['updated']:
            self.last_update = newlocal['updated']
        return newlocal

    @staticmethod
    def poll():  # pragma: no cover
        """Check the data provider for updates."""

        raise NotImplementedError

    @staticmethod
    def event_proto_to_local(proto):  # pragma: no cover
        """Convert an event from native data provider format to standardized local format."""

        raise NotImplementedError

    @staticmethod
    def event_local_to_proto(local):  # pragma: no cover
        """Convert an event from standardized local format to native data provider format."""

        raise NotImplementedError

    @staticmethod
    def proto_add(proto):  # pragma: no cover
        """Ask the data provider to create an event."""

        raise NotImplementedError

    @staticmethod
    def proto_remove(proto_id):  # pragma: no cover
        """Ask the data provider to remove an event."""

        raise NotImplementedError

    @staticmethod
    def proto_update(proto_id, proto):  # pragma: no cover
        """Ask the data provider to update an event."""

        raise NotImplementedError



"""A manager that blends multiple base.Calendar objects' events together."""

import operator
import time

from metabot.calendars import loader


class MultiCalendar:
    """A manager that blends multiple base.Calendar objects' events together."""

    _current_index = None

    def __init__(self):
        self.calendars = {}
        self.by_local_id = {}
        self.ordered = []

    def _rebuild(self):
        self.ordered.sort(key=operator.itemgetter('start', 'end', 'summary', 'local_id'))
        current_index = None
        by_local_id = {}
        now = time.time()
        for i, event in enumerate(self.ordered):
            by_local_id[event['local_id']] = i
            if current_index is None and event['end'] >= now:
                current_index = i
        self._current_index = current_index
        self.by_local_id = by_local_id

    @property
    def current_index(self):
        """The index into self.ordered of the current (earliest ongoing or next to begin) event."""

        if self._current_index is None:
            return
        now = time.time()
        while self.ordered[self._current_index]['end'] < now:
            if self._current_index == len(self.ordered) - 1:
                self._current_index = None
                break
            self._current_index += 1
        return self._current_index

    @property
    def current_local_id(self):
        """The local_id of the current event."""

        cur = self.current_index
        if cur is not None:
            return self.ordered[cur]['local_id']

    def add(self, calid):
        """Add a new calendar to the manager (if it's not already installed)."""

        if calid not in self.calendars:
            self.calendars[calid] = loader.get(calid)
            self.ordered.extend(self.calendars[calid].events.values())
            self._rebuild()
        return self.calendars[calid]

    def get_event(self, local_id=None):
        """Retrieve a specific event, plus the event immediately before and after it."""

        if local_id is None:
            index = self.current_index
        else:
            index = self.by_local_id.get(local_id)
        if index is None:
            return None, None, None
        prevev = nextev = None
        if index > 0:
            prevev = self.ordered[index - 1]
        if index < len(self.ordered) - 1:
            nextev = self.ordered[index + 1]
        return prevev, self.ordered[index], nextev

    def get_overlap(self, start, end):
        """Find all events whose [start, end] range overlaps the given [start, end] range."""

        for event in self.ordered:
            if event['start'] > end:
                break
            if event['end'] >= start:
                yield event

    def poll(self):
        """Poll all installed calendars for updates."""

        updated = False
        for calendar in self.calendars.values():
            updated = calendar.poll() or updated
        if updated:
            self.ordered = []
            for calendar in self.calendars.values():
                self.ordered.extend(calendar.events.values())
            self._rebuild()
        return updated

    def view(self, calcodes):
        """A MultiCalendar-like object that operates on a subset of the installed calendars."""

        return View(self, calcodes)


class View:
    """A MultiCalendar-like object that operates on a subset of the installed calendars."""

    def __init__(self, multical, calcodes):
        self.multical = multical
        self.calcodes = calcodes

    @property
    def current_index(self):
        """The index into self.multical.ordered of the current event."""

        cur = self.multical.current_index
        if cur is None:
            return
        while self.multical.ordered[cur]['local_id'].split(':', 1)[0] not in self.calcodes:
            if cur == len(self.multical.ordered) - 1:
                return
            cur += 1
        return cur

    @property
    def current_local_id(self):
        """The local_id of the current event."""

        cur = self.current_index
        if cur is not None:
            return self.multical.ordered[cur]['local_id']

    def get_event(self, local_id=None):
        """Retrieve a specific event, plus the event immediately before and after it."""

        if local_id is None:
            index = self.current_index
        else:
            index = self.multical.by_local_id.get(local_id)
        if index is None:
            return None, None, None
        prevev = nextev = None

        previndex = index - 1
        while previndex >= 0:
            prevev = self.multical.ordered[previndex]
            if prevev['local_id'].split(':', 1)[0] in self.calcodes:
                break
            prevev = None
            previndex -= 1

        nextindex = index + 1
        while nextindex <= len(self.multical.ordered) - 1:
            nextev = self.multical.ordered[nextindex]
            if nextev['local_id'].split(':', 1)[0] in self.calcodes:
                break
            nextev = None
            nextindex += 1

        return prevev, self.multical.ordered[index], nextev

    def get_overlap(self, start, end):
        """Find all events whose [start, end] range overlaps the given [start, end] range."""

        for event in self.multical.get_overlap(start, end):
            if event['local_id'].split(':', 1)[0] in self.calcodes:
                yield event
